import inspect

from umk import core
from umk.core.typings import Callable, Any, Type


class Defer:
    def __init__(self, func=None, **kwargs):
        self.func: Callable[..., Any] | None = func
        self.args: dict[str, Any] = kwargs

    def __bool__(self):
        return self.func is not None

    def __call__(self, min: int, max: int, _0: Any = None, _1: Any = None, _2: Any = None):
        return call(self.func, min, max, _0, _1, _2)


class SourceError(core.Error):
    def __init__(self, *messages: str, details: core.Properties = None):
        super().__init__(name=type(self).__name__.rstrip("Error"))
        self.messages = list(messages)
        self.details = details or self.details


class FunctionError(core.Error):
    def __init__(self, *messages: str, details: core.Properties = None):
        super().__init__(name=type(self).__name__.rstrip("Error"))
        self.messages = list(messages)
        self.details = details or self.details


class ClassError(core.Error):
    def __init__(self, *messages: str, details: core.Properties = None):
        super().__init__(name=type(self).__name__.rstrip("Error"))
        self.messages = list(messages)
        self.details = details or self.details


class SubclassError(core.Error):
    def __init__(self, *messages: str, details: core.Properties = None):
        super().__init__(name=type(self).__name__.rstrip("Error"))
        self.messages = list(messages)
        self.details = details or self.details


class ExistsError(core.Error):
    def __init__(self, *messages: str, details: core.Properties = None):
        super().__init__(name=type(self).__name__.rstrip("Error"))
        self.messages = list(messages)
        self.details = details or self.details


class NotFoundError(core.Error):
    def __init__(self, *messages: str, details: core.Properties = None):
        super().__init__(name=type(self).__name__.rstrip("Error"))
        self.messages = list(messages)
        self.details = details or self.details


class SignatureError(core.Error):
    def __init__(self, *messages: str, details: core.Properties = None):
        super().__init__(name=type(self).__name__.rstrip("Error"))
        self.messages = list(messages)
        self.details = details or self.details


class RequirementError(core.Error):
    def __init__(self, *messages: str, details: core.Properties = None):
        super().__init__(name=type(self).__name__.rstrip("Error"))
        self.messages = list(messages)
        self.details = details or self.details


class NotRegisteredError(core.Error):
    def __init__(self, *messages: str, details: core.Properties = None):
        super().__init__(name=type(self).__name__.rstrip("Error"))
        self.messages = list(messages)
        self.details = details or self.details


def call(func, min: int, max: int, _0: Any = None, _1: Any = None, _2: Any = None):
    sig = len(inspect.signature(func).parameters)
    if sig < min:
        core.globals.error_console.print(
            f"Failed to call '{func.__name__}': arguments valid range is \[min={min}], but {sig} required"
        )
        core.globals.close(-1)
    if max > -1 and sig > max:
        core.globals.error_console.print(
            f"Failed to call '{func.__name__}': arguments valid range is \[min={min} max={max}], but {sig} required"
        )
        core.globals.close(-1)
    if sig == 0:
        return func()
    if sig == 1:
        return func(_0)
    elif sig == 2:
        return func(_0, _1)
    elif sig == 3:
        return func(_0, _1, _2)


# def cpf(func):
#     """Based on http://stackoverflow.com/a/6528148/190597 (Glenn Maynard)"""
#     if inspect.isclass(func):
#         return func
#     result = types.FunctionType(
#         code=func.__code__,
#         globals=func.__globals__,
#         name=func.__name__,
#         argdefs=func.__defaults__,
#         closure=func.__closure__
#     )
#     result = functools.update_wrapper(result, func)
#     result.__kwdefaults__ = func.__kwdefaults__
#     return result


class Decorator(core.Model):
    class OnErrors(core.Model):
        module: core.Error = core.Field(
            default_factory=core.Error,
            description="Error on invalid module"
        )
        subject: core.Error = core.Field(
            default_factory=core.Error,
            description="Error on invalid subject"
        )
        single: core.Error = core.Field(
            default_factory=core.Error,
            description="Error on multiple usage"
        )
        sig: core.Error = core.Field(
            default_factory=core.Error,
            description="Error on invalid signature"
        )
        base: core.Error = core.Field(
            default_factory=core.Error,
            description="Error on invalid base type of the input class"
        )

    class Input(core.Model):
        class Signature(core.Model):
            min: int = core.Field(
                default=0,
                description="Minimum arguments count"
            )
            max: int = core.Field(
                default=-1,
                description="Maximum arguments count"
            )
        subject: str = core.Field(
            default="any",
            description="Which subject must be used this decorator with (class/function/any)"
        )
        base: tuple[Type] | Type | None = core.Field(
            default=None,
            description="Input class base type"
        )
        sig: Signature = core.Field(
            default_factory=Signature,
            description="Signature constraints"
        )

    defers: list[Defer] = core.Field(
        default_factory=list,
        description="Defer function"
    )
    stack: int = core.Field(
        default=2,
        description="Stack level to validate on"
    )
    module: str = core.Field(
        default="",
        description="Which module this decorator is allowed to use in"
    )
    target: str = core.Field(
        default="",
        description="Framework target function name"
    )
    errors: OnErrors = core.Field(
        default_factory=OnErrors,
        description="Errors for each case"
    )
    single: bool = core.Field(
        default=False,
        description="Reject multiple registration"
    )
    skip: bool = core.Field(
        default=False,
        description="Skip registration"
    )
    input: Input = core.Field(
        default_factory=Input,
        description="Input function/class constraints"
    )
    registered: bool = core.Field(
        default=False,
        description="Is register was called ?"
    )

    def init(self, **kwargs):
        pass

    def validate(self, f):
        frame = inspect.stack()[self.stack]
        script = f".unimake/{self.module}.py"
        if not frame.filename.endswith(script):
            raise self.errors.module
        if self.single and self.registered:
            raise self.errors.single
        if self.input.subject == "class":
            if not inspect.isclass(f):
                raise self.errors.subject
            if self.input.base and not issubclass(f, self.input.base):
                raise self.errors.base
        elif self.subject == "function":
            if not inspect.isfunction(f):
                raise self.errors.subject
        sig = len(inspect.signature(f).parameters)
        if sig < self.sig.min:
            raise self.errors.signature
        if self.sig.max > 0:
            if sig > self.sig.max:
                raise self.errors.sig

    @core.overload
    def register(self, f): ...

    @core.overload
    def register(self, f, **kwargs): ...

    def register(self, f=None, **kwargs):
        if f is not None:
            if not self.skip:
                # self.defers.append(Defer(func=cpf(f)))
                self.defers.append(Defer(func=f))
                self.registered = True
            return f

        def dec(fun):
            # parse kwargs
            if not self.skip:
                # self.defers.append(Defer(func=cpf(fun), **kwargs))
                self.defers.append(Defer(func=fun, **kwargs))
                self.registered = True
            return fun

        return dec



from umk import core
from umk.framework import utils
from umk.framework.filesystem import AnyPath, OptPath
from umk.framework.system.environs import OptEnv


class Interface(core.Model):
    name: str = core.Field(
        default="",
        description="Remote environment name",
    )
    description: str = core.Field(
        default="",
        description="Remote environment description",
    )
    default: bool = core.Field(
        default=False,
        description="Whether this remote environment are default or not",
    )

    def object(self) -> core.Object:
        result = core.Object()
        result.type = "Remote." + type(self).__name__.split(".")[-1]
        for name in self.model_fields:
            props = self.property(name)
            for prop in props:
                result.properties.add(prop)
        return result

    def property(self, name: str) -> list[core.Property]:
        field = self.model_fields[name]
        if not field.repr:
            return []
        return [
            core.Property(
                name=name,
                description=field.description,
                value=getattr(self, name)
            )
        ]

    def build(self, **kwargs):
        """
        Build remote environment.
        """
        self.__not_implemented()

    def destroy(self, **kwargs):
        """
        Destroy remote environment.
        """
        self.__not_implemented()

    def up(self, **kwargs):
        """
        Start remote environment.
        """
        self.__not_implemented()

    def down(self, **kwargs):
        """
        Stop remote environment.
        """
        self.__not_implemented()

    def shell(self, **kwargs):
        """
        Open remote environment shell.
        """
        self.__not_implemented()

    def login(self, **kwargs):
        """
        Login remote environment.
        """
        self.__not_implemented()

    def logout(self, **kwargs):
        """
        Logout remote environment.
        """
        self.__not_implemented()

    @core.typeguard
    def execute(self, cmd: list[AnyPath], cwd: OptPath = None, env: OptEnv = None, **kwargs):
        """
        Execute command in remote environment.
        """
        self.__not_implemented()

    @core.typeguard
    def upload(self, items: dict[AnyPath, AnyPath], **kwargs):
        """
        Upload given paths to remote environment.
        """
        self.__not_implemented()

    @core.typeguard
    def download(self, items: dict[AnyPath, AnyPath], **kwargs):
        """
        Download given paths from remote environment.
        """
        self.__not_implemented()

    def __not_implemented(self):
        core.globals.console.print(
            f"[bold]The '{self.name}' remote environment has no '{utils.caller(2)}' function. "
            f"It`s must be managed outside of this tool."
        )

    def __hash__(self) -> int:
        return hash(self.name)

    def __str__(self):
        return f"umk.remote.Interface(name='{self.name}', description='{self.description}')"



from umk import core
from umk.kit import config, target, project
from umk.core.typings import Generator
from umk.runtime import utils


class Targets(core.Model):
    class Decorators(core.Model):
        custom: utils.Decorator = core.Field(
            description="Decorator of the target 'custom'",
            default_factory=lambda: utils.Decorator(
                stack=2,
                module="targets",
                input=utils.Decorator.Input(
                    subject="class",
                    base=target.Interface,
                ),
                errors=utils.Decorator.OnErrors(
                    module=utils.SourceError("Failed to register custom target outside of the .unimake/targets.py"),
                    subject=utils.ClassError("Failed to register custom target. Use 'umk.framework.targets.custom' with classes based on 'umk.framework.targets.Interface'"),
                    base=utils.SubclassError("Failed to register custom target. Use 'umk.framework.targets.custom' with classes based on 'umk.framework.targets.Interface'"),
                )
            )
        )
        function: utils.Decorator = core.Field(
            description="Decorator of the target 'function'",
            default_factory=lambda: utils.Decorator(
                stack=2,
                input=utils.Decorator.Input(
                    subject="function",
                ),
                module="targets",
                errors=utils.Decorator.OnErrors(
                    module=utils.SourceError("Failed to register target 'function' outside of the .unimake/targets.py"),
                    subject=utils.FunctionError("Failed to register target 'function'. Use 'umk.framework.targets.function' with functions")
                )
            ),
        )
        go_binary: utils.Decorator = core.Field(
            description="Decorator of the target 'go.binary'",
            default_factory=lambda: utils.Decorator(
                stack=2,
                input=utils.Decorator.Input(
                    subject="function",
                    sig=utils.Decorator.Input.Signature(min=1)
                ),
                module="targets",
                errors=utils.Decorator.OnErrors(
                    module=utils.SourceError("Failed to register target 'go.binary' outside of the .unimake/targets.py"),
                    subject=utils.FunctionError("Failed to register target 'go.binary'. Use 'umk.framework.targets.go.binary' with functions"),
                    sig=utils.SignatureError("Failed to register target 'go.binary'. Function must accept 1 argument at least"),
                )
            ),
        )
        go_mod: utils.Decorator = core.Field(
            description="Decorator of the target 'go.mod'",
            default_factory=lambda: utils.Decorator(
                stack=2,
                input=utils.Decorator.Input(
                    subject="function",
                    sig=utils.Decorator.Input.Signature(min=1)
                ),
                module="targets",
                errors=utils.Decorator.OnErrors(
                    module=utils.SourceError("Failed to register target 'go.mod' outside of the .unimake/targets.py"),
                    subject=utils.FunctionError("Failed to register target 'go.mod'. Use 'umk.framework.targets.go.binary' with functions"),
                    sig=utils.SignatureError("Failed to register target 'go.mod'. Function must accept 1 argument at least"),
                )
            ),
        )
        command: utils.Decorator = core.Field(
            description="Decorator of the target 'command'",
            default_factory=lambda: utils.Decorator(
                stack=2,
                input=utils.Decorator.Input(
                    subject="function",
                    sig=utils.Decorator.Input.Signature(min=1)
                ),
                module="targets",
                errors=utils.Decorator.OnErrors(
                    module=utils.SourceError("Failed to register target 'command' outside of the .unimake/targets.py"),
                    subject=utils.FunctionError("Failed to register target 'command'. Use 'umk.framework.targets.go.binary' with functions"),
                    sig=utils.SignatureError("Failed to register target 'command'. Function must accept 1 argument at least"),
                )
            ),
        )
        packages: utils.Decorator = core.Field(
            description="Decorator of the target 'packages'",
            default_factory=lambda: utils.Decorator(
                stack=2,
                input=utils.Decorator.Input(
                    subject="function",
                    sig=utils.Decorator.Input.Signature(min=1)
                ),
                module="targets",
                errors=utils.Decorator.OnErrors(
                    module=utils.SourceError("Failed to register target 'packages' outside of the .unimake/targets.py"),
                    subject=utils.FunctionError(
                        "Failed to register target 'command'. Use 'umk.framework.targets.packages' with functions"),
                    sig=utils.SignatureError(
                        "Failed to register target 'packages'. Function must accept 1 argument at least"),
                )
            ),
        )
    decorator: Decorators = core.Field(
        default_factory=Decorators,
        description="Targets decorators"
    )
    items: dict[str, target.Interface] = core.Field(
        default_factory=dict,
        description="Project targets"
    )

    def __iter__(self):
        for t in self.items.values():
            yield t

    def __contains__(self, name: str):
        return name in self.items

    def objects(self) -> Generator[target.Interface, None, None]:
        for item in self.items.values():
            yield item.object()

    def get(self, name: str, on_err=None) -> target.Interface:
        return self.items.get(name, on_err)

    def run(self, *names: str):
        found = []
        for name in names:
            if name not in self.items:
                core.globals.console.print(f"[yellow bold]Target '{name}' not found")
            else:
                found.append(name)
        for name in found:
            t = self.get(name)
            p = project.get()
            c = config.get()
            t.run(p=p, c=c)

    def init(self):
        target.run = self.run
        target.function = self.decorator.function.register
        target.custom = self.decorator.custom.register
        target.command = self.decorator.command.register
        target.packages = self.decorator.packages.register
        target.go.binary = self.decorator.go_binary.register
        target.go.mod = self.decorator.go_mod.register

    def setup(self, c: config.Interface, p: project.Interface):
        def append(tar: target.Interface):
            if tar.name in self.items:
                e = utils.ExistsError()
                e.messages.append(f"Target '{tar.name}' is already registered")
                e.details.new(name="name", value=tar.name, desc="Target name")
                raise e
            self.items[tar.name] = tar

        for defer in self.decorator.function.defers:
            t = target.Function(
                name=defer.args.get("name", defer.func.__name__),
                description=defer.args.get("description", (defer.func.__doc__ or "").strip()),
                label=defer.args.get("label", ""),
                function=defer.func
            )
            append(t)
        for defer in self.decorator.go_binary.defers:
            src = target.GolangBinary()
            sig = self.decorator.go_binary.input.sig
            with_debug = defer.args.get("debug", True)
            defer(sig.min, sig.max, src, c, p)
            if with_debug:
                d, r = target.GolangBinary.new(
                    name=src.name,
                    label=src.label,
                    description=src.description,
                    tool=src.tool,
                    build=src.build,
                    port=src.debug.port,
                )
                append(d)
                append(r)
            else:
                append(src)
        for defer in self.decorator.command.defers:
            src = target.Command()
            sig = self.decorator.go_binary.input.sig
            defer(sig.min, sig.max, src, c, p)
            append(src)
        for defer in self.decorator.go_mod.defers:
            src = target.GolangMod()
            sig = self.decorator.go_mod.input.sig
            defer(sig.min, sig.max, src, c, p)
            append(src)
        for defer in self.decorator.packages.defers:
            src = target.SystemPackages()
            sig = self.decorator.packages.input.sig
            defer(sig.min, sig.max, src, c, p)
            append(src)
        for defer in self.decorator.custom.defers:
            res = defer(0, 2, c, p)
            append(res)



import copy
from umk.core.typings import Any

from umk import core
from umk.framework.utils import caller
from umk.framework.filesystem import Path


class Opt(core.Model):
    name: str = core.Field(default="", description="CLI option (or flag) qualified name")

    def list(self) -> list[str]: ...


class Str(Opt):
    value: str = core.Field(default="", description="Option value")
    surr: str = core.Field(default="", description="Surrounded symbols")
    equal: str = core.Field(default="", description="CLI option (flag) equal symbol (--opt=val, --opt val, ...)")

    def list(self) -> list[str]:
        self.surr = self.surr.strip()
        if self.equal == "":
            return [self.name, f'{self.surr}{self.value}{self.surr}']
        else:
            return [f"{self.name}{self.equal}{self.surr}{self.value}{self.surr}"]


class File(Opt):
    value: Path = core.Field(default="", description="Option value")
    equal: str = core.Field(default="", description="CLI option (flag) equal symbol (--opt=val, --opt val, ...)")

    def list(self) -> list[str]:
        if self.equal == "":
            return [self.name, f'{self.value.as_posix()}']
        else:
            return [f"{self.name}{self.equal}{self.value.as_posix()}"]


class Int(Opt):
    value: int = core.Field(default=0, description="Option value")
    surr: str = core.Field(default="", description="Surrounded symbols")
    equal: str = core.Field(default="", description="CLI option (flag) equal symbol (--opt=val, --opt val, ...)")

    def list(self) -> list[str]:
        self.surr = self.surr.strip()
        if self.equal == "":
            return [self.name, f'{self.surr}{self.value}{self.surr}']
        else:
            return [f"{self.name}{self.equal}{self.surr}{self.value}{self.surr}"]


class Bool(Opt):
    value: bool = core.Field(default=False, description="Option value")
    surr: str = core.Field(default="", description="Surrounded symbols")
    kind: str = core.Field(default="", description="Represent style (alpha, digit, word)")
    equal: str = core.Field(default="", description="CLI option (flag) equal symbol (--opt=val, --opt val, ...)")

    def list(self) -> list[str]:
        self.surr = self.surr.strip()
        quoted = lambda string: f"{self.surr}{string}{self.surr}"
        if self.kind == "digit":
            return [self.name, quoted(str(int(self.value)))]
        elif self.kind == "word":
            return [self.name, quoted({True: "yes", False: "no"}[self.value])]
        return {True: [self.name], False: []}[self.value]


class List(Opt):
    value: list = core.Field(default_factory=list, description="Option value")
    split: str = core.Field(default="")
    surr: str = core.Field(default="", description="Surrounded symbols of the whole result")
    surr_each: str = core.Field(default="", description="Surrounded symbols of the each item")
    multi: bool = core.Field(default=False, description="Use option for each item (--opt=item1 --opt=item2 ...)")
    equal: str = core.Field(default="", description="CLI option (flag) equal symbol (--opt=val, --opt val, ...)")

    def list(self) -> list[str]:
        sa = self.surr.strip()
        se = self.surr_each.strip()
        eq = self.equal if self.equal else " "
        sp = self.split if self.split else " "
        chunks = []
        for value in self.value:
            value = f'{se}{value}{se}'
            if self.multi:
                if eq.strip():
                    chunk = [f"{self.name}{eq}{value}"]
                else:
                    chunk = [self.name, value]
            else:
                chunk = [value]
            chunks += chunk
        if not self.multi:
            return [f"{self.name}{eq}{sa}{sp.join(chunks)}{sa}"]
        return chunks


class Dict(Opt):
    value: dict = core.Field(default_factory=dict, description="Option value")
    split: str = core.Field(default=" ")
    surr: str = core.Field(default="", description="Surrounded symbols of the whole result")
    surr_each: str = core.Field(default="", description="Surrounded symbols of the each item")
    multi: bool = core.Field(default=False, description="Use option for each item (--opt=item1 --opt=item2 ...)")
    equal: str = core.Field(default="", description="CLI option (flag) equal symbol (--opt=val, --opt val, ...)")
    equal_each: str = core.Field(default="=", description="Value equal symbol (--opt=k:v, --opt=k=v, ...)")
    equal_each_skip: bool = core.Field(default=True, description="Skip equal symbol when value is empty")

    def list(self) -> list[str]:
        sa = self.surr.strip()
        se = self.surr_each.strip()
        eq = self.equal if self.equal else " "
        ee = self.equal_each if self.equal_each else " "
        sp = self.split if self.split else " "
        chunks = []
        for k, v in self.value.items():
            if v:
                item = f'{se}{k}{ee}{v}{se}'
            else:
                item = f'{se}{k}{se}'
            if self.multi:
                if eq.strip():
                    chunk = [f"{self.name}{eq}{item}"]
                else:
                    chunk = [self.name, item]
            else:
                chunk = [item]
            chunks += chunk
        if not self.multi:
            return [f"{self.name}{eq}{sa}{sp.join(chunks)}{sa}"]
        return chunks


class Obj(Opt):
    value: Any = core.Field(default=None, description="Option value")
    surr: str = core.Field(default="", description="Surrounded symbols")
    equal: str = core.Field(default="", description="CLI option (flag) equal symbol (--opt=val, --opt val, ...)")

    def list(self) -> list[str]:
        self.surr = self.surr.strip()
        res = str(self.value).lower()
        if self.equal == "":
            return [self.name, f'{self.surr}{res}{self.surr}']
        else:
            return [f"{self.name}{self.equal}{self.surr}{res}{self.surr}"]


class Arg(Opt):
    value: Any = core.Field(default=None, description="Argument value")
    # required: bool = core.Field(default=False, description="Whether argument is required or not")

    def list(self) -> list[str]:
        if self.value is not None:
            return [str(self.value)]
        return []


class Args(Opt):
    value: list[Any] = core.Field(default=None, description="Argument values")

    def list(self) -> list[str]:
        result = []
        for entry in self.value:
            if entry is not None:
                result.append(str(entry))
        return result


class NoEmpty(core.Model):
    class Config(core.Model.Config):
        excluder = core.field.empty


class Options(NoEmpty):
    def serialize(self) -> list[str]:
        result = []
        for name in self.model_fields:
            f = self.model_fields[name]
            cli: None | Opt | Arg | Args = f.json_schema_extra.get("cli") if f.json_schema_extra else None
            if not cli:
                result.extend(getattr(self, name).serialize())
            elif issubclass(type(cli), Opt):
                items = self._opt(cli, name)
                result.extend(items)
            elif issubclass(type(cli), Arg):
                items = self._arg(cli, name)
                result.extend(items)
            elif issubclass(type(cli), Args):
                items = self._args(cli, name)
                result.extend(items)
            else:
                items = self._other(cli, name)
                result.extend(items)
        return result

    def _opt(self, opt: Opt, field: str) -> list[str]:
        value = getattr(self, field)
        if issubclass(type(value), (list, dict, set)) and not value or value is None:
            return []
        o = copy.deepcopy(opt)
        o.value = value
        return o.list()

    def _arg(self, arg: Arg, field: str) -> list[str]:
        value = getattr(self, field)
        if issubclass(type(value), (list, dict, set)) and not value or value is None:
            return []
        a = copy.deepcopy(arg)
        a.value = value
        return a.list()

    def _args(self, args: Args, field: str) -> list[str]:
        value = getattr(self, field)
        if issubclass(type(value), (list, dict, set)) and not value or value is None:
            return []
        a = copy.deepcopy(args)
        a.value = value
        return a.list()

    def _other(self, cli: Any, field: str) -> list[str]:
        core.globals.print(f"[bold yellow] {caller()}: invalid cli option type, name='{field}' type='{type(cli)}'")
        return []



import sys

from umk import core
from umk.framework.adapters import docker
from umk.framework.filesystem import AnyPath, OptPath
from umk.framework.remote.interface import Interface
from umk.framework.system.environs import OptEnv
from umk.framework.system.shell import Shell
from umk.framework.system.user import User


class Login(core.Model):
    server: str = core.Field(default="", description="Server URL")
    user: str = core.Field(default="", description="Docker repository user")
    password: str = core.Field(default="", description="Docker repository password")


class Compose(Interface):
    service: str = core.Field(
        default=None,
        description="Target compose service"
    )
    sh: list[str] = core.Field(
        default_factory=lambda: ["bash"],
        description="Shell command (sh, bash, zsh, ...)"
    )
    tty: bool = core.Field(
        default_factory=lambda: sys.stdout.isatty(),
        description="Instantiate tty when call 'execute'"
    )
    composefile: None | docker.ComposeFile = core.Field(
        default_factory=docker.ComposeFile,
        description="Docker compose file object."
    )
    dockerfiles: list[docker.File] = core.Field(
        default_factory=list,
        description="Dockerfile objects."
    )
    logins: list[Login] = core.Field(
        default_factory=list,
        description="Private repositories login info"
    )

    @property
    def client(self) -> docker.Client:
        return docker.Client(
            compose_files=[self.composefile.file]
        )

    @core.typeguard
    def build(self, *args, **kwargs):
        # save docker files
        for dockerfile in self.dockerfiles:
            dockerfile.save()

        # save compose file
        self.composefile.save()

        services = list(self.composefile.services.keys())

        # clear old
        self.client.compose.down(remove_orphans=True, remove_images="all", volumes=True, quiet=False)
        self.client.compose.rm(services=services, stop=True, volumes=True)

        # build images
        self.client.compose.build(services=services)

        # create containers
        self.client.compose.create(services=services, no_recreate=True)

    def destroy(self, *args, **kwargs):
        services = list(self.composefile.services.keys())
        self.client.compose.down(remove_orphans=True, remove_images="all", volumes=True, quiet=False)
        self.client.compose.rm(services=services, stop=True, volumes=True)

    def login(self, **kwargs):
        if not self.logins:
            core.globals.console.print(
                "[bold]No private repositories to login to !"
            )
            return
        for info in self.logins:
            core.globals.console.print(f"Docker login: '{info.server}'")
            self.client.login(info.server, info.user, info.password)

    @core.typeguard
    def up(self, *args, **kwargs):
        services = list(self.composefile.services.keys())
        self.client.compose.up(services=services, build=False, detach=True, remove_orphans=True)

    def down(self, *args, **kwargs):
        self.client.compose.down(remove_orphans=True, volumes=True, quiet=False)

    def shell(self, **kwargs):
        service = kwargs.get("service", self.service)
        self.client.compose.execute(
            service=service,
            command=self.sh,
            detach=False,
            tty=True,
        )

    @core.typeguard
    def execute(self, cmd: list[AnyPath], cwd: OptPath = None, env: OptEnv = None, **kwargs):
        detach = kwargs.get("detach") or False
        self.client.compose.execute(
            service=self.service,
            command=cmd,
            envs=env or {},
            tty=self.tty,
            workdir=cwd,
            detach=detach
        )

    @core.typeguard
    def upload(self, items: dict[AnyPath, AnyPath], **kwargs):
        service = kwargs.get("service", self.service)
        for i, (src, dst) in enumerate(items.items()):
            shell = Shell(cmd=self.client.compose.docker_compose_cmd)
            shell.cmd += ["cp", str(src), f"{service}:{dst}"]
            shell.sync(log=False)

    @core.typeguard
    def download(self, items: dict[AnyPath, AnyPath], **kwargs):
        service = kwargs.get("service", self.service)
        for i, (src, dst) in enumerate(items.items()):
            shell = Shell(cmd=self.client.compose.docker_compose_cmd)
            shell.cmd += ["cp", f"{service}:{src}", str(dst)]
            shell.sync(log=False)

    def property(self, name: str) -> list[core.Property]:
        if name != "dockerfiles":
            return super().property(name)
        return [
            core.Property(
                name=f"dockerfile[{i}]",
                description=f"Dockerfile object [{i}]",
                value=df
            ) for i, df in enumerate(self.dockerfiles, start=0)
        ]


class Container(Interface):
    container: str = core.Field(
        default="",
        description="Target container name",
    )
    sh: list[str] = core.Field(
        default_factory=lambda: ["bash"],
        description="Shell command (sh, bash, zsh, ...)"
    )
    workdir: OptPath = core.Field(
        default="/",
        description="Shell working directory."
    )
    environments: OptEnv = core.Field(
        default=None,
        description="Shell environment variables."
    )
    privileged: bool = core.Field(
        default=False,
        description="Open privileged shell."
    )
    user: User | None = core.Field(
        default=None,
        description="Open shell by user."
    )

    @property
    def client(self) -> docker.Client:
        return docker.Client()

    def shell(self, *args, **kwargs):
        usr = None
        if self.user:
            usr = f"{self.user.id}:{self.user.group.id}"
        self.client.container.execute(
            container=self.container,
            command=self.sh,
            detach=False,
            envs=self.environments or {},
            interactive=True,
            privileged=self.privileged,
            tty=True,
            user=usr,
            workdir=self.workdir,
            stream=False
        )

    @core.typeguard
    def execute(self, cmd: list[AnyPath], cwd: OptPath = None, env: OptEnv = None, **kwargs):
        envs = {}
        if self.environments:
            envs.update(self.environments)
        if env:
            envs.update(env)
        if not envs:
            envs = None
        usr = None
        if self.user:
            usr = f"{self.user.id}:{self.user.group.id}"
        detach = kwargs.get("detach") or False
        self.client.container.execute(
            container=self.container,
            command=self.sh + cmd,
            detach=detach,
            envs=envs,
            interactive=False,
            privileged=self.privileged,
            tty=False,
            user=usr,
            workdir=self.workdir,
            stream=False
        )

    @core.typeguard
    def upload(self, items: dict[AnyPath, AnyPath], **kwargs):
        for i, (src, dst) in items.items():
            self.client.container.copy(
                source=src,
                destination=(self.container, dst)
            )

    @core.typeguard
    def download(self, items: dict[AnyPath, AnyPath], **kwargs):
        for i, (src, dst) in items.items():
            self.client.container.copy(
                source=(self.container, src),
                destination=dst
            )



import abc
import asyncio
import subprocess
import sys
from asyncio import subprocess as async_subprocess
from pathlib import Path
from umk.core.typings import Callable

from umk import core
from umk.framework.system.environs import Environs

Command = str | list[str]


class Handler:
    @abc.abstractmethod
    def on_error(self, text: str): ...

    @abc.abstractmethod
    def on_output(self, text: str): ...

    @abc.abstractmethod
    def on_exception(self, exc: Exception): ...


class Devnull(Handler):
    def on_error(self, text: str): ...

    def on_output(self, text: str): ...

    def on_exception(self, exc: Exception): ...


class Colorful(Handler, core.Model):
    out: str = core.Field(default="[bold green]${msg}", description="Output pattern")
    err: str = core.Field(default="[bold red]${msg}", description="Error pattern")
    exc: str = core.Field(default="[bold red]${exc}", description="Error pattern")

    def on_error(self, text: str):
        for line in text.split('\n'):
            if line.strip():
                core.globals.console.print(self.out.replace("${msg}", line))

    def on_output(self, text: str):
        for line in text.split('\n'):
            if line.strip():
                core.globals.console.print(self.err.replace("${msg}", line))

    def on_exception(self, exc: Exception):
        core.globals.console.print(self.err.replace("${exc}", str(exc)))


class Fetch(Handler, core.Model):
    out: list[str] = core.Field(default_factory=list, description="Output buffer")
    err: list[str] = core.Field(default_factory=list, description="Error buffer")
    exc: None | Exception = core.Field(default=None, description="Exception object")

    def on_error(self, text: str):
        self.err.append(text)

    def on_output(self, text: str):
        self.out.append(text)

    def on_exception(self, exc: Exception):
        self.exc = exc

    def outstr(self, splitter="\n") -> str:
        return splitter.join(self.out)

    def errstr(self, splitter="\n") -> str:
        return splitter.join(self.err)

    def excstr(self) -> str:
        return str(self.exc) if self.exc else ""


class Std(Handler):
    def on_error(self, text: str): ...

    def on_output(self, text: str): ...

    def on_exception(self, exc: Exception):
        core.globals.log.error(msg=str(exc))


devnull = subprocess.DEVNULL
pipe = subprocess.PIPE
stdout = sys.stdout
stderr = sys.stderr


class Shell(core.Model):
    name: str = core.Field(
        default="",
        description="Shell name (just for convenience)"
    )
    cmd: list[str | Path] = core.Field(
        default_factory=list,
        description="Shell command and it's options"
    )
    workdir: None | Path | str = core.Field(
        default=None,
        description="Shell working directory"
    )
    environs: None | Environs = core.Field(
        default=None,
        description="Shell environment variables"
    )
    handler: None | Handler = core.Field(
        default_factory=Std,
        description="Result handler"
    )
    log: bool = core.Field(
        default=False,
        description="Print command or not"
    )
    stringifier: Callable[[list[str | Path]], str] = core.Field(
        default=lambda args: " ".join([str(entry) for entry in args]),
        description="List to string converter (it's need to convert command list to string)"
    )

    async def asyn(self, *, log: bool | None = None) -> int | None:
        cmd = self.stringifier(self.cmd)
        self._log_cmd(log, cmd)
        inp, out, err = self._descriptors()
        prc: async_subprocess.Process | None = None
        try:
            prc = await asyncio.create_subprocess_shell(
                cmd=cmd,
                stdin=inp,
                stdout=out,
                stderr=err,
                cwd=self.workdir,
                env=self.environs,
                shell=True
            )
        except Exception as e:
            if self.handler:
                self.handler.on_exception(e)
            return

        if out != pipe and err != pipe:
            return await prc.wait()

        async def read(stream, func):
            while True:
                line = await stream.readline()
                if not line:
                    return
                func(self.handler, line.rstrip().decode("utf-8"))

        await asyncio.gather(
            read(prc.stdout, type(self.handler).on_output),
            read(prc.stderr, type(self.handler).on_error)
        )

        return await prc.wait()

    def sync(self, *, log: bool | None = None) -> int | None:
        self._log_cmd(log, self.stringifier(self.cmd))
        prc: subprocess.Popen | None = None
        inp, out, err = self._descriptors()

        try:
            prc = subprocess.Popen(
                args=self.cmd,
                stdout=out,
                stderr=err,
                universal_newlines=True,
                cwd=self.workdir,
                env=self.environs,
                shell=False
            )
        except Exception as e:
            self.handler.on_exception(e)
            return

        if out != err != pipe:
            return prc.wait()

        while True:
            code = prc.poll()
            if code is not None:
                if prc.stdout:
                    for line in prc.stdout.readlines():
                        self.handler.on_output(line.rstrip())
                if prc.stderr:
                    for line in prc.stderr.readlines():
                        self.handler.on_error(line.rstrip())
                return code
            if prc.stdout:
                o = prc.stdout.readline()
                if o:
                    self.handler.on_output(o.rstrip())
            if prc.stderr:
                e = prc.stderr.readline()
                if e:
                    self.handler.on_error(e.strip())

    def _descriptors(self):
        inp = None
        out = stdout
        err = stderr
        if self.handler:
            t = type(self.handler)
            if issubclass(t, Devnull):
                out = devnull
                err = devnull
            elif issubclass(t, Std):
                out = stdout
                err = stderr
            else:
                out = pipe
                err = pipe
        return inp, out, err

    def _log_cmd(self, need: bool | None, cmd: str):
        need_log = need if need is not None else self.log
        if need_log:
            if self.name:
                core.globals.console.print(f"[bold]shell\['{self.name}']: {cmd}")
            else:
                core.globals.console.print(f"[bold]shell: {cmd}")



# Dockerfile
from umk.framework.adapters.docker.file import File
from umk.framework.adapters.docker.file import Add as FileAdd
from umk.framework.adapters.docker.file import Arg as FileArg
from umk.framework.adapters.docker.file import Cmd as FileCmd
from umk.framework.adapters.docker.file import Copy as FileCopy
from umk.framework.adapters.docker.file import Entrypoint as FileEntrypoint
from umk.framework.adapters.docker.file import Env as FileEnv
from umk.framework.adapters.docker.file import Expose as FileExpose
from umk.framework.adapters.docker.file import From as FileFrom
from umk.framework.adapters.docker.file import Healthcheck as FileHealthcheck
from umk.framework.adapters.docker.file import Label as FileLabel
from umk.framework.adapters.docker.file import Maintainer as FileMaintainer
from umk.framework.adapters.docker.file import OnBuild as FileOnBuild
from umk.framework.adapters.docker.file import Run as FileRun
from umk.framework.adapters.docker.file import Shell as FileShell
from umk.framework.adapters.docker.file import User as FileUser
from umk.framework.adapters.docker.file import Volume as FileVolume
from umk.framework.adapters.docker.file import Workdir as FileWorkdir

# Compose
from umk.framework.adapters.docker.compose import BlockIo as ComposeBlockIo
from umk.framework.adapters.docker.compose import Build as ComposeBuild
from umk.framework.adapters.docker.compose import Config as ComposeConfig
from umk.framework.adapters.docker.compose import Credential as ComposeCredential
from umk.framework.adapters.docker.compose import Dependency as ComposeDependency
from umk.framework.adapters.docker.compose import Deploy as ComposeDeploy
from umk.framework.adapters.docker.compose import Develop as ComposeDevelop
from umk.framework.adapters.docker.compose import Device as ComposeDevice
from umk.framework.adapters.docker.compose import EnvFile as ComposeEnvFile
from umk.framework.adapters.docker.compose import Extend as ComposeExtend
from umk.framework.adapters.docker.compose import File as ComposeFile
from umk.framework.adapters.docker.compose import Healthcheck as ComposeHealthcheck
from umk.framework.adapters.docker.compose import IPAM as ComposeIpam
from umk.framework.adapters.docker.compose import IpamConfig as ComposeIpamConfig
from umk.framework.adapters.docker.compose import Logging as ComposeLogging
from umk.framework.adapters.docker.compose import Mount as ComposeMount
from umk.framework.adapters.docker.compose import Net as ComposeNet
from umk.framework.adapters.docker.compose import Network as ComposeNetwork
from umk.framework.adapters.docker.compose import Placement as ComposePlacement
from umk.framework.adapters.docker.compose import Resource as ComposeResource
from umk.framework.adapters.docker.compose import Resources as ComposeResources
from umk.framework.adapters.docker.compose import Restart as ComposeRestart
from umk.framework.adapters.docker.compose import Secret as ComposeSecret
from umk.framework.adapters.docker.compose import SecretAccess as ComposeSecretAccess
from umk.framework.adapters.docker.compose import Service as ComposeService
from umk.framework.adapters.docker.compose import StorageOpt as ComposeStorageOpt
from umk.framework.adapters.docker.compose import ULimits as ComposeULimits
from umk.framework.adapters.docker.compose import Update as ComposeUpdate
from umk.framework.adapters.docker.compose import Volume as ComposeVolume
from umk.framework.adapters.docker.compose import Volumes as ComposeVolumes
from umk.framework.adapters.docker.compose import Watch as ComposeWatch


__all__ = [
    "File",
    "FileAdd",
    "FileArg",
    "FileCmd",
    "FileCopy",
    "FileEntrypoint",
    "FileEnv",
    "FileExpose",
    "FileFrom",
    "FileHealthcheck",
    "FileLabel",
    "FileMaintainer",
    "FileOnBuild",
    "FileRun",
    "FileShell",
    "FileUser",
    "FileVolume",
    "FileWorkdir",

    "ComposeBlockIo",
    "ComposeBuild",
    "ComposeConfig",
    "ComposeCredential",
    "ComposeDependency",
    "ComposeDeploy",
    "ComposeDevelop",
    "ComposeDevice",
    "ComposeEnvFile",
    "ComposeExtend",
    "ComposeFile",
    "ComposeHealthcheck",
    "ComposeIpam",
    "ComposeIpamConfig",
    "ComposeLogging",
    "ComposeMount",
    "ComposeNet",
    "ComposeNetwork",
    "ComposePlacement",
    "ComposeResource",
    "ComposeResources",
    "ComposeRestart",
    "ComposeSecret",
    "ComposeSecretAccess",
    "ComposeService",
    "ComposeStorageOpt",
    "ComposeULimits",
    "ComposeUpdate",
    "ComposeVolume",
    "ComposeVolumes",
    "ComposeWatch",
]



from pathlib import Path

from fs.base import FS
from fs.osfs import OSFS
from fs.walk import Walker
from fs import move as _mv

from umk import core


@core.overload
def move(
    src: str | Path,
    dst: str | Path,
):
    """
    Moves file or directory to destination path (on local filesystem).
    It will create all destination sub-dirs if not exists.

    :param src: Path to file or directory to move from.
    :param dst: Path to file or directory to move to.

    Examples
    ::
     - move(Path('./main.py'), Path('./hello.py'))
     - move(Path('./hellp.py'), Path('./some/dir/hello.py'))
    """
    s = OSFS(src.parent.as_posix())
    d = OSFS(dst.parent.as_posix(), create=True)
    _mv.move_file(s, src.name, d, dst.name)


@core.overload
def move(
    src: str | Path,
    dst: tuple[FS, str]
):
    """
    Moves specific local file/directory to destination filesystem.
    It will create all destination sub-dirs if not exists.

    :param src: Source file/directory path
    :param dst: Tuple of filesystem to move to and path to file/directory in destination filesystem

    Examples
     - move('./some/file.txt', (ftp(..), 'description.ini'))
    """
    raise NotImplemented()


@core.overload
def move(
    src: tuple[FS, str],
    dst: str | Path
):
    """
    Moves specific file or directory from source filesystem to local file.
    It will create all destination sub-dirs if not exists.

    :param src: Tuple of filesystem to move from and path to file/directory in source filesystem
    :param dst: Destination file/directory path

    Examples
     - move((ftp(..), 'description.ini'), './some/file.txt')
    """
    raise NotImplemented()


@core.overload
def move(
    src: tuple[FS, str],
    dst: tuple[FS, str]
):
    """
    Moves specific file or directory from source FS to destination FS.
    It will create all destination sub-dirs if not exists.

    :param src: Tuple of filesystem to move from and path to file/directory in source filesystem
    :param dst: Tuple of filesystem to move to and path to file/directory in destination filesystem

    Examples
     - move(local(..), 'some/file.txt', ftp(..), 'file.txt')
    """
    raise NotImplemented()


@core.overload
def move(
    src: FS,
    dst: FS,
    wlk: Walker
):
    """
    Moves walker items from source FS to destination FS.
    It will create all destination sub-dirs if not exists.

    :param src: Filesystem to move from
    :param dst: Filesystem to move to
    :param wlk: Walker from source filesystem

    Examples:
     - move(local(..), ftp(..), walker(..))
    """
    raise NotImplemented()



import platform as bp

from umk import core


class Platform(core.Model):
    os: str = core.Field(
        default="",
        description="OS name"
    )
    arch: str = core.Field(
        default="",
        description="Architecture"
    )

    @property
    def unix(self) -> bool:
        return self.os in ("linux", "darwin")

    @property
    def linux(self) -> bool:
        return self.os == "linux"

    @property
    def windows(self) -> bool:
        return self.os == "windows"

    @property
    def x64(self) -> bool:
        return self.arch in ("amd64", "x86_64", "arm64")


def platform() -> Platform:
    result = Platform()
    result.os = bp.system().lower()
    if result.os == "win32":
        result.os = "windows"
    result.arch = bp.machine()
    return result



import os
import sys


class RemoteState:
    def __init__(self, completion: bool):
        self.default: bool = False
        self.specific: str = ""
        self.cmd = []

        if completion:
            return

        if len(sys.argv) < 1:
            return

        commands = {
            "run": True,
            "release": True,
            "config": False,
            "inspect": False,
            "format": False,
            "remote": False,
            "target": False,
        }

        position = None
        for i, arg in enumerate(sys.argv[1:]):
            if arg in commands:
                return
            if arg.startswith("-R="):
                self.specific = arg.lstrip("-R=")
                position = i
                break
            elif arg == "-R":
                self.default = True
                position = i
                break

        if position is None:
            return

        subcmd = ""
        allowed = False
        for arg in sys.argv[1:]:
            if arg in commands:
                subcmd = arg
                allowed = commands.get(arg)
                break
        if not subcmd:
            self.default = False
            self.specific = ""
            return

        if not allowed:
            print(f"Subcommand '{subcmd}' is not remotable !", file=sys.stderr)
            sys.exit(-1)
        else:
            sys.argv.pop(position + 1)
            self.cmd = ["umk"] + sys.argv[1:]


complete: bool = "_UMK_COMPLETE" in os.environ
unsafe: bool = "UMK_UNSAFE" in os.environ
remote = RemoteState(complete)



import beartype
import pydantic as pd
from beartype.typing import *
from beartype import beartype as typeguard
from multimethod import overload as base_overload
from pydantic import Field


beartype.BeartypeConf.is_color = False

Adapter = pd.TypeAdapter
SerializationInfo = pd.SerializationInfo
overload = base_overload


class model:
    serializer = pd.model_serializer

    @staticmethod
    @typeguard
    def dict(obj: pd.BaseModel, auxiliary=None) -> dict[str, Any]:
        if obj is None:
            return auxiliary
        return obj.model_dump()

class field:
    serializer = pd.field_serializer
    validator = pd.field_validator

    @staticmethod
    def empty(value: Any) -> bool:
        t = type(value)
        if issubclass(t, bool):
            return False
        return not bool(value)


class Model(pd.BaseModel):
    class Config:
        arbitrary_types_allowed = True
        validate_assignment = True

    @model.serializer(mode="wrap")
    def serialize_model(self, wrap: Callable[[...], dict[str, Any]], info: pd.SerializationInfo) -> Any:
        result = wrap(self)
        for n, f in self.model_fields.items():
            if f.exclude:
                result.pop(n, None)
            v = getattr(self, n)
            e = self.model_config.get("excluder") or (f.json_schema_extra or {}).get("excluder")
            if e and e(v):
                result.pop(n, None)
        return result



from umk import core
from umk.kit import project
from umk.kit import config
from umk.core.typings import Type, Callable, Any
from umk.runtime import utils


class EntryDecorator(utils.Decorator):
    klass: Type | None = core.Field(
        default=None,
        description="Project type to pass to entry function (Scratch, Golang, ...)"
    )


class Project(core.Model):
    class Decorators(core.Model):
        release: utils.Decorator = core.Field(
            description="Decorator of the project release function",
            default_factory=lambda: utils.Decorator(
                stack=2,
                input=utils.Decorator.Input(
                    subject="function",
                ),
                module="project",
                single=True,
                errors=utils.Decorator.OnErrors(
                    module=utils.SourceError("Failed to register releaser outside of the .unimake/project.py"),
                    subject=utils.FunctionError("Failed to register release. Use 'project.releaser' with functions")
                )
            ),
        )
        entry: EntryDecorator = core.Field(
            description="Decorator of the project empty, golang, custom, ... functions",
            default_factory=lambda: EntryDecorator(
                stack=2,
                module="project",
                single=True,
                errors=utils.Decorator.OnErrors(
                    module=utils.SourceError("Failed to register project outside of the .unimake/project.py"),
                    single=utils.ExistsError("Failed to register project. It's already exists"),
                )
            )
        )
    decorator: Decorators = core.Field(
        default_factory=Decorators,
        description="Project decorators"
    )
    releaser: Callable[..., Any] | None = core.Field(
        default=None,
        description="Project release function"
    )
    instance: project.Interface = core.Field(
        default_factory=project.Interface,
        description="Project instance"
    )

    def release(self):
        if self.releaser is None:
            core.globals.console.print(
                f"[bold]Project '{self.instance.info.name or self.instance.info.id}' has no [/][cyan bold]release[/] action !"
            )
        else:
            utils.call(self.releaser, 0, 2, config.get(), self.instance)

    def entry(self, factory, target: str = "custom", klass: Type | None = None):
        self.decorator.entry.target = target
        self.decorator.entry.klass = klass
        if target == "custom":
            self.decorator.entry.input.subject = "class"
            self.decorator.entry.input.sig.min = 0
            self.decorator.entry.errors.subject = utils.ClassError(
                f"Failed to register project. Use 'umk.framework.project.custom' with functions"
            )
            self.decorator.entry.errors.base = utils.SubclassError(
                "Failed to register project. Entry must be based on umk.framework.project.Interface"
            )
        else:
            self.decorator.entry.input.subject = "function"
            self.decorator.entry.input.sig.min = 1
            self.decorator.entry.errors.sig = utils.SignatureError(
                f"Failed to register project. Function must accept project instance"
            )
            self.decorator.entry.errors.subject = utils.FunctionError(
                f"Failed to register project. Use 'umk.framework.project.{self.decorator.entry.target} with functions"
            )
        return self.decorator.entry.register(factory)

    def init(self):
        project.get = lambda: self.instance
        project.release = self.release
        project.releaser = self.decorator.release.register
        project.empty = lambda f: self.entry(f, "empty", project.Scratch)
        project.golang = lambda f: self.entry(f, "golang", project.Golang)
        project.custom = lambda f: self.entry(f)

    def setup(self, c: config.Interface):
        if not self.decorator.entry.registered:
            raise utils.NotRegisteredError(
                "No project registration was found. Put one in the .unimake/project.py'"
            )
        if self.decorator.entry.klass is None:
            self.instance = self.decorator.entry.defers[0](0, 1, c)
        else:
            self.instance = self.decorator.entry.klass()
            self.decorator.entry.defers[0](1, 2, self.instance, c)
        if self.decorator.release.registered:
            if not self.instance:
                raise utils.NotRegisteredError(
                    "Failed to register project releaser. No project registration was found. Put one in the .unimake/project.py'"
                )
            self.releaser = self.decorator.release.defers[0].func
        self.decorator.entry.skip = True
        self.decorator.release.skip = True



import sys
from importlib import util as importer
from pathlib import Path

from umk import core
from umk.kit import config
from umk.kit import remote
from umk.runtime.config import Config
from umk.runtime.project import Project
from umk.runtime.targets import Targets
from umk.runtime.remotes import Remote


class Options(core.Model):
    class Config(core.Model):
        overrides: dict[str, config.Value] = core.Field(
            default_factory=dict,
            description="Config entry overrides (passed by unimake CLI)"
        )
        presets: list[str] = core.Field(
            default_factory=list,
            description="Config presets to apply (passed by unimake CLI)"
        )
        file: bool = core.Field(
            default=False,
            description="Load config from file and modify it by preset and overrides"
        )

    root: Path = core.Field(
        default=core.globals.paths.unimake,
        description="Path to '.unimake' directory"
    )
    config: Config = core.Field(
        default_factory=Config,
        description="Config options"
    )


class Container:
    def __init__(self):
        self._modules = {}
        self.config = Config()
        self.project = Project()
        self.targets = Targets()
        self.remotes = Remote()

    def load(self, options: Options):
        root = options.root.expanduser().resolve().absolute()
        sys.path.insert(0, root.as_posix())

        with_config = (root / "config.py").exists()
        with_remote = (root / "remote.py").exists()

        self.config.init()
        self.project.init()
        self.targets.init()
        self.remotes.init()

        if with_config:
            self.script(root, "config")
        self.script(root, "project")
        if with_remote:
            self.script(root, "remote")

        if with_config:
            self.config.setup(options.config)
        self.project.setup(self.config.instance)
        self.targets.setup(self.config.instance, self.project.instance)
        if with_remote:
            self.remotes.setup(self.config.instance, self.project.instance)

    def find_remote(self, default: bool, specific: str) -> remote.Interface:
        if default:
            result = self.remotes.get(self.remotes.default)
            if not result:
                core.globals.error_console.print(
                    'Failed to find default remote environment! '
                    'Please specify it in the .unimake/remote.py'
                )
                core.globals.close(-1)
            return result
        if specific:
            result = self.remotes.get(specific)
            if not result:
                core.globals.error_console.print(
                    f"Failed to find remote environment '{specific}'! "
                    f"Please create it in the .unimake/remote.py"
                )
                core.globals.close(-1)
            return result

    def script(self, root: Path, name: str):
        file = root / f'{name}.py'
        spec = importer.spec_from_file_location(name, file)
        module = importer.module_from_spec(spec)
        sys.modules[f'umk:{name}'] = module
        spec.loader.exec_module(module)
        self._modules[name] = module




import os

import asyncclick

from umk.application.cmd import root
from umk.application import utils

if not os.environ.get('_UMK_COMPLETE', None):
    from umk import runtime, core
    from umk.application.printers import PropertiesPrinter


@root.group(help="Config management commands")
def config():
    pass


@config.command(help="Remove saved config file")
def clean():
    runtime.c.config.clean()


@config.command(help="Save project config")
@utils.options.config.preset
@utils.options.config.entry
def save(c: tuple[str], p: tuple[str]):
    opt = runtime.Options()
    opt.config = utils.config(False, p, c)
    runtime.c.load(opt)
    runtime.c.config.save()


@config.command(help="Write entry inside config file")
@asyncclick.argument('values', required=True, nargs=-1)
def write(values: tuple[str]):
    opt = runtime.Options()
    opt.config.file = True
    runtime.c.load(opt)
    entries = utils.config(False, None, values).overrides
    runtime.c.config.write(entries)


@config.command(help="Print config presets")
@utils.options.style
def presets(s: str):
    opt = runtime.Options()
    runtime.c.load(opt)

    if not runtime.c.config.presets:
        core.globals.console.print(f"[bold]Config presets not found !")
        return

    data = {name: func.__doc__ or "" for name, func in runtime.c.config.presets.items()}
    if s == "style":
        properties = core.Properties()
        for name, desc in data.items():
            properties.new(name=name, desc=desc, value=None)
        printer = PropertiesPrinter()
        printer.print(properties, value=False)
    elif s == "json":
        core.globals.console.print_json(
            json=core.json.text(data)
        )


@config.command(name='inspect', help="Print default config details")
@utils.options.style
@utils.options.config.all
def inspect(s: str, c: tuple[str], p: tuple[str], f: bool):
    opts = runtime.Options()
    opts.config = utils.config(f, p, c)
    runtime.c.load(opts)

    conf = runtime.c.config
    if not conf.instance:
        core.globals.console.print(f"[bold]Config: config not found, register one at first !")
        return

    data = conf.object()
    if s == "style":
        printer = PropertiesPrinter()
        printer.print(data.properties)
    elif s == "json":
        core.globals.console.print_json(
            json=core.json.text(data)
        )



from umk.framework.remote.docker import Compose as DockerCompose
from umk.framework.remote.docker import Container as DockerContainer
from umk.framework.remote.docker import Login as DockerLogin
from umk.framework.remote.interface import Interface
from umk.framework.remote.ssh import SecureShell

__all__ = [
    "Interface",
    "DockerContainer",
    "DockerCompose",
    "DockerLogin",
    "SecureShell",
]


class docker:
    @staticmethod
    def compose(factory):
        # See implementation in runtime.Instance.implementation()
        raise NotImplemented()

    @staticmethod
    def container(factory):
        # See implementation in runtime.Instance.implementation()
        raise NotImplemented()


def ssh(factory):
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


def custom():
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


def find(name: str, on_err=None):
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


def default(on_err=None):
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


def iterate():
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()



import abc
import inspect

from umk import core
from umk.core.typings import Callable, Any
from umk.framework.system.shell import Shell


class Interface(core.Model):
    name: str = core.Field(
        default="",
        description="Target name"
    )
    label: str = core.Field(
        default="",
        description="Target label"
    )
    description: str = core.Field(
        default="",
        description="Target description"
    )

    def object(self, **kwargs) -> core.Object:
        result = core.Object()
        result.type = "Target"
        result.properties.new(name="Name", value=self.name, desc="Target name")
        result.properties.new(name="Label", value=self.label, desc="Target label")
        result.properties.new(name="Description", value=self.description, desc="Target description")
        return result

    @abc.abstractmethod
    def run(self, **kwargs):
        raise NotImplemented()


class Command(Interface):
    shell: Shell = core.Field(
        default_factory=Shell,
        description="Shell to run"
    )

    def run(self, **kwargs):
        if self.shell.cmd:
            self.shell.sync()

    def object(self) -> core.Object:
        result = super().object()
        result.type = "Target.Command"
        result.properties.new("Command", self.shell.cmd, "Shell command")
        result.properties.new("Workdir", self.shell.workdir or "", "Working directory")
        result.properties.new("Environs", self.shell.environs or "", "Environment variables")
        return result


class Function(Interface):
    function: Callable[..., Any] = core.Field(
        default=None,
        description="Function to run"
    )

    def run(self, **kwargs):
        if self.function:
            sig = len(inspect.signature(self.function).parameters)
            if sig == 0:
                self.function()
            elif sig == 1:
                self.function(kwargs.get("c"))
            else:
                self.function(kwargs.get("c"), kwargs.get("p"))

    def object(self) -> core.Object:
        result = super().object()
        result.type = "Target.Function"
        return result





import os

import asyncclick

from umk.application import utils

if not os.environ.get('_UMK_COMPLETE'):
    from umk import runtime, core
    from rich.table import Table
    from umk.framework.system.shell import Shell


@asyncclick.group()
@utils.options.remote
async def root(r):
    pass


@root.command(help="Run project targets")
@utils.options.config.all
@asyncclick.argument('names', required=True, nargs=-1)
def run(c: tuple[str], p: tuple[str], f: bool, names: tuple[str]):
    opt = runtime.Options()
    opt.config = utils.config(f, p, c)
    runtime.c.load(opt)
    utils.forward(runtime.c)
    runtime.c.targets.run(*names)


@root.command(name='inspect', help="Inspect project details")
@utils.options.style
@utils.options.config.all
def inspect(s: str, c: tuple[str], p: tuple[str], f: bool):
    opt = runtime.Options()
    opt.config = utils.config(f, p, c)
    runtime.c.load(opt)

    pro = runtime.c.project.instance
    tar = runtime.c.targets
    if s in ("style", ""):
        table_info = Table(
            title="INFO",
            title_style="bold cyan",
            title_justify="left",
            show_header=False,
            show_edge=False,
            show_lines=False,
            box=None,
        )
        table_info.add_column("Name", justify="left", style="bold", no_wrap=True)
        table_info.add_column("Value", justify="left", style="", no_wrap=True)
        table_info.add_row("• Id", pro.info.id)
        table_info.add_row("• Version", pro.info.version)
        table_info.add_row("• Name", pro.info.name)
        table_info.add_row("• Description", pro.info.description)

        if pro.info.contributors:
            table_contributors = Table(
                title="CONTRIBUTORS",
                title_style="bold cyan",
                title_justify="left",
                show_header=False,
                show_edge=False,
                show_lines=False,
                box=None,
            )
            table_contributors.add_column("Name", justify="left", style="bold", no_wrap=True)
            table_contributors.add_column("Value", justify="left", style="", no_wrap=True)
            for contrib in pro.info.contributors:
                contacts = " ".join(contrib.email)
                contacts += " " + " ".join([f'{k}:{v}' for k, v in contrib.socials.items()])
                table_contributors.add_row(f"• {contrib.name}", contacts)

        if tar:
            table_targets = Table(
                title="TARGETS",
                title_style="bold cyan",
                title_justify="left",
                show_header=False,
                show_edge=False,
                show_lines=False,
                box=None,
            )
            table_targets.add_column("Name", justify="left", style="bold", no_wrap=True)
            table_targets.add_column("Description", justify="left", style="", no_wrap=True)
            for t in tar:
                table_targets.add_row(f"• {t.name}", t.description)
        core.globals.console.print(table_info)
        if tar:
            core.globals.console.print()
            core.globals.console.print(table_targets)
        if pro.info.contributors:
            core.globals.console.print()
            core.globals.console.print(table_contributors)
    elif s == "json":
        # TODO Append targets
        core.globals.console.print_json(core.json.text(pro.info))


@root.command(help="Release project")
@utils.options.config.all
def release(c: tuple[str], p: tuple[str], f: bool):
    opt = runtime.Options()
    opt.config = utils.config(f, p, c)
    runtime.c.load(opt)
    utils.forward(runtime.c)
    runtime.c.project.release()


@root.command(name='format', help="Format .unimake/*.py files")
def prettier():
    if not core.globals.paths.unimake.exists():
        core.globals.console.print("[bold]No '.unimake' folder was found")
    else:
        sh = Shell()
        sh.name = "format"
        sh.cmd = ["black", "-l", "100", "-t", "py311", "-W", "4", core.globals.paths.unimake]
        sh.sync()



import os

if not os.environ.get('_UMK_COMPLETE', None):
    from rich.table import Table
    from umk import core


    class PropertiesPrinter:
        def print(self, *entries: core.Properties, name=True, description=True, value=True, splitter=" ", equal="="):
            for properties in entries:
                table = Table(show_header=True, show_edge=True, show_lines=True)
                if name:
                    table.add_column("Name", justify="left", style="", no_wrap=True)
                if description:
                    table.add_column("Description", justify="left", style="italic", no_wrap=True)
                if value:
                    table.add_column("Value", justify="left", style="", no_wrap=True)
                for item in properties:
                    if isinstance(item.value, list):
                        val = splitter.join(item.value)
                    elif isinstance(item.value, dict):
                        val = "\n".join([f"{k}{equal}{v}" for k, v in item.value.items()])
                    else:
                        val = str(item.value)
                    row = []
                    if name:
                        row.append(item.name)
                    if description:
                        row.append(item.description or "<empty>")
                    if value:
                        row.append(val)
                    table.add_row(*row)
                core.globals.console.print(table)



import copy

from umk import core
from umk.framework.filesystem import Path
from umk.framework.adapters.go import Go as Tool
from umk.framework.adapters.go import Build as GoBuild
from umk.framework.target.interface import Interface


class GolangBinary(Interface):
    class Debug(core.Model):
        port: int = core.Field(
            default=2345,
            description="Port to run delve on"
        )

    tool: Tool = core.Field(
        default_factory=Tool,
        description="Golang tool object"
    )
    build: GoBuild = core.Field(
        default_factory=GoBuild,
        description="Build options"
    )
    debug: Debug = core.Field(
        default_factory=Debug,
        description="Debug options"
    )

    @staticmethod
    @core.typeguard
    def new(*, name: str, tool: Tool, build: GoBuild, port: int = 2345, label: str = "", description: str = "") -> tuple['GolangBinary', 'GolangBinary']:
        base = GolangBinary(
            name=name.strip(),
            label=label.strip(),
            description=description.strip(),
            tool=tool,
            build=build,
            debug=GolangBinary.Debug(port=port)
        )
        if not base.label:
            base.label = f"Binary '{base.name}'"
        if not base.description:
            base.description = f"Golang binary ({base.name})"

        debug = copy.deepcopy(base)
        debug.name = name
        debug.label += " (debug)"
        debug.build.flags.gc.append('all=-N')
        debug.build.flags.gc.append('-l')
        if debug.description:
            debug.description += " (with debug info)"

        release = copy.deepcopy(base)
        release.name = name + ".release"
        release.label += " (release)"
        release.build.flags.gc.append('-dwarf=false')
        release.build.flags.ld.append('-s')
        release.build.flags.ld.append('-w')

        return debug, release

    def object(self) -> core.Object:
        result = super().object()
        result.type = "Target.Golang.Binary"
        result.properties.new("Tool", self.tool.shell.cmd, "Golang tool object")
        result.properties.new("Build", " ".join(self.build.serialize()), "Build options")
        return result

    def run(self, **kwargs):
        self.tool.build(self.build)


class GolangMod(Interface):
    tool: Tool = core.Field(
        default_factory=Tool,
        description="Go tool object"
    )
    path: Path | None = core.Field(
        default=None,
        description="Path to directory with go.mod"
    )
    compat: str = core.Field(
        default="",
        description="Preserves any additional checksums (see 'go help mod tidy')"
    )
    vendor: bool = core.Field(
        default=False,
        description="Vendors downloaded packages"
    )

    def run(self, **kwargs):
        self.tool.shell.workdir = self.path
        self.tool.mod.tidy(compat=self.compat)
        if self.vendor:
            self.tool.mod.vendor()

    def object(self) -> core.Object:
        result = core.Object()
        result.type = "Target.Golang.Mod"
        result.properties.new("Tool", self.tool.shell.cmd, "Golang tool object")
        result.properties.new("Path", self.path, "Directory with go.mod")
        result.properties.new("Compat", self.compat, desc="Preserves any additional checksums (see 'go help mod tidy')")
        result.properties.new("Vendor", self.vendor, "Vendors packages or not")
        return result




from pathlib import Path

from umk.framework.filesystem.copy import copy as cp
from umk.framework.filesystem.move import move as mv
from umk.framework.filesystem.factories import generic, memory, local, zip, ftp, tmp, sub

AnyPath = Path | str
OptPath = Path | str | None



import copy

from umk import core
from umk.framework.utils import cli
from umk.framework.filesystem import Path, AnyPath
from umk.framework.system.shell import Shell


class BuildOptions(cli.Options):
    @staticmethod
    def new(mode: str, output: AnyPath, *sources: AnyPath) -> 'BuildOptions':
        match mode:
            case "release":
                return BuildOptions.release(output, *sources)
            case "debug":
                return BuildOptions.debug(output, *sources)
        raise ValueError(f"Invalid mode for golang build options: given={mode}, expect=[release, debug]")

    @staticmethod
    def release(output: AnyPath, *sources: AnyPath) -> 'BuildOptions':
        result = BuildOptions()
        result.output = output
        result.flags.gc.append('-dwarf=false')
        result.flags.ld.append('-s')
        result.flags.ld.append('-w')
        if sources:
            result.source = sources
        return result

    @staticmethod
    def debug(output: AnyPath, *sources: AnyPath) -> 'BuildOptions':
        result = BuildOptions()
        result.output = output
        result.flags.gc.append('all=-N')
        result.flags.gc.append('-l')
        if sources:
            result.source = sources
        return result

    class Print(cli.Options):
        packages: bool = core.Field(
            default=False,
            cli=cli.Bool(name="-v"),
            description="Print the names of packages as they are compiled."
        )
        commands: bool = core.Field(
            default=False,
            cli=cli.Bool(name="-x"),
            description="Print the commands."
        )

    class Cover(cli.Options):
        enabled: None | bool = core.Field(
            default=None,
            cli=cli.Bool(name="-cover"),
            description="Enable code coverage instrumentation."
        )
        mode: None | str = core.Field(
            default=None,
            cli=cli.Str(name="-covermode"),
            description="Set the mode for coverage analysis.the default is 'set' unless -race is enabled,in which case it is 'atomic'.the values:set: bool: does this statement run?count: int: how many times does this statement run?atomic: int: count, but correct in multithreaded tests;significantly more expensive.sets -cover."
        )
        pkg: list[str] = core.Field(
            default_factory=list,
            cli=cli.List(name="-coverpkg"),
            description="For a build that targets package 'main' (e.g. building a goexecutable), apply coverage analysis to each package matchingthe patterns. the default is to apply coverage analysis topackages in the main go module. see 'go help packages' for adescription of package patterns.  sets -cover."
        )

    class Profiling(cli.Options):
        race: None | bool = core.Field(
            default=None,
            cli=cli.Bool(name="-race"),
            description="Enable data race detection.supported only on linux/amd64, freebsd/amd64, darwin/amd64, darwin/arm64, windows/amd64,linux/ppc64le and linux/arm64 (only for 48-bit vma)."
        )
        memory: None | bool = core.Field(
            default=None,
            cli=cli.Bool(name="-msan"),
            description="Enable interoperation with memory sanitizer.supported only on linux/amd64, linux/arm64, freebsd/amd64and only with clang/llvm as the host c compiler.pie build mode will be used on all platforms except linux/amd64."
        )
        address: None | bool = core.Field(
            default=None,
            cli=cli.Bool(name="-asan"),
            description="Enable interoperation with address sanitizer.supported only on linux/arm64, linux/amd64.supported only on linux/amd64 or linux/arm64 and only with gcc 7 and higheror clang/llvm 9 and higher."
        )

    class Mod(cli.Options):
        strategy: None | str = core.Field(
            default=None,
            cli=cli.Str(name="-mod"),
            description="Module download mode to use: readonly, vendor, or mod.by default, if a vendor directory is present and the go version in go.modis 1.14 or higher, the go command acts as if -mod=vendor were set.otherwise, the go command acts as if -mod=readonly were set.see https://golang.org/ref/mod#build-commands for details."
        )
        cacherw: None | bool = core.Field(
            default=None,
            cli=cli.Bool(name="-modcacherw"),
            description="Leave newly-created directories in the module cache read-writeinstead of making them read-only."
        )
        file: None | Path | str = core.Field(
            default=None,
            cli=cli.Str(name="-modfile"),
            description="In module aware mode, read (and possibly write) an alternate go.modfile instead of the one in the module root directory. a file named'go.mod' must still be present in order to determine the module rootdirectory, but it is not accessed. when -modfile is specified, analternate go.sum file is also used: its path is derived from the-modfile flag by trimming the '.mod' extension and appending '.sum'."
        )

    class Flags(cli.Options):
        gc: list[str] = core.Field(
            default_factory=list,
            cli=cli.List(name="-gcflags", surr="", equal="="),
            description="Arguments to pass on each go tool compile invocation."
        )
        ld: list[str] = core.Field(
            default_factory=list,
            cli=cli.List(name="-ldflags", surr="", equal="="),
            description="Arguments to pass on each go tool link invocation."
        )
        gccgo: list[str] = core.Field(
            default_factory=list,
            cli=cli.List(name="-gccgoflags", surr="", equal="="),
            description="Arguments to pass on each gccgo compiler/linker invocation."
        )
        asm: list[str] = core.Field(
            default_factory=list,
            cli=cli.List(name="-asmflags", surr="", equal="="),
            description="Arguments to pass on each go tool asm invocation."
        )

    output: None | Path | str = core.Field(
        default=None,
        cli=cli.File(name="-o"),
        description="Binary file output."
    )
    force: bool = core.Field(
        default=False,
        cli=cli.Bool(name="-a"),
        description="Force rebuilding of packages that are already up-to-date."
    )
    dry: None | bool = core.Field(
        default=None,
        cli=cli.Bool(name="-n"),
        description="Print the commands but do not run them."
    )
    job: None | int = core.Field(
        default=None,
        cli=cli.Int(name="-p"),
        description="The number of programs, such as build commands ortest binaries, that can be run in parallel.the default is gomaxprocs, normally the number of cpus available."
    )
    print: Print = core.Field(
        default_factory=Print,
        description="Print some info during building."
    )
    cover: Cover = core.Field(
        default_factory=Cover,
        description="Set the mode for coverage analysis.the default is 'set' unless -race is enabled,in which case it is 'atomic'.the values:set: bool: does this statement run?count: int: how many times does this statement run?atomic: int: count, but correct in multithreaded tests;significantly more expensive.sets -cover."
    )
    flags: Flags = core.Field(
        default_factory=Flags,
        description="Compiler, linker, gccgo and asm flags."
    )
    profiling: Profiling = core.Field(
        default_factory=Profiling,
        description="Profiling options."
    )
    work: None | bool = core.Field(
        default=None,
        cli=cli.Bool(name="-work"),
        description="Print the name of the temporary work directory anddo not delete it when exiting."
    )
    buildmode: None | str = core.Field(
        default=None,
        cli=cli.Str(name="-buildmode"),
        description="Build mode to use. see 'go help buildmode' for more."
    )
    buildvcs: None | bool = core.Field(
        default=None,
        cli=cli.Bool(name="-buildvcs"),
        description="Whether to stamp binaries with version control information('true', 'false', or 'auto'). by default ('auto'), version controlinformation is stamped into a binary if the main package, the main modulecontaining it, and the current directory are all in the same repository.use -buildvcs=false to always omit version control information, or-buildvcs=true to error out if version control information is available butcannot be included due to a missing tool or ambiguous directory structure."
    )
    compiler: None | str = core.Field(
        default=None,
        cli=cli.Str(name="-compiler"),
        description="Name of compiler to use, as in runtime.compiler (gccgo or gc)."
    )
    mod: Mod = core.Field(
        default_factory=Mod,
        description="Mod options."
    )
    installsuffix: None | str = core.Field(
        default=None,
        cli=cli.Str(name="-installsuffix"),
        description="A suffix to use in the name of the package installation directory,in order to keep output separate from default builds.if using the -race flag, the install suffix is automatically set to raceor, if set explicitly, has _race appended to it. likewise for the -msanand -asan flags. using a -buildmode option that requires non-default compileflags has a similar effect."
    )
    linkshared: None | bool = core.Field(
        default=None,
        cli=cli.Bool(name="-linkshared"),
        description="Build code that will be linked against shared libraries previouslycreated with -buildmode=shared."
    )
    overlay: None | Path | str = core.Field(
        default=None,
        cli=cli.Str(name="-overlay"),
        description="Read a json config file that provides an overlay for build operations.the file is a json struct with a single field, named 'replace', thatmaps each disk file path (a string) to its backing file path, so thata build will run as if the disk file path exists with the contentsgiven by the backing file paths, or as if the disk file path does notexist if its backing file path is empty. support for the -overlay flaghas some limitations: importantly, cgo files included from outside theinclude path must be in the same directory as the go package they areincluded from, and overlays will not appear when binaries and tests arerun through go run and go test respectively."
    )
    pgo: None | Path | str = core.Field(
        default=None,
        cli=cli.Str(name="-pgo"),
        description="Specify the file path of a profile for profile-guided optimization (pgo).when the special name 'auto' is specified, for each main package in thebuild, the go command selects a file named 'default.pgo' in the package'sdirectory if that file exists, and applies it to the (transitive)dependencies of the main package (other packages are not affected).special name 'off' turns off pgo. the default is 'auto'."
    )
    pkgdir: None | Path | str = core.Field(
        default=None,
        cli=cli.Str(name="-pkgdir"),
        description="Install and load all packages from dir instead of the usual locations.for example, when building with a non-standard configuration,use -pkgdir to keep generated packages in a separate location."
    )
    tags: list[str] = core.Field(
        default_factory=list,
        cli=cli.List(name="-tags"),
        description="A comma-separated list of additional build tags to consider satisfiedduring the build. for more information about build tags, see'go help buildconstraint'. (earlier versions of go used aspace-separated list, and that form is deprecated but still recognized.)"
    )
    trimpath: None | bool = core.Field(
        default=None,
        cli=cli.Bool(name="-trimpath"),
        description="Remove all file system paths from the resulting executable.instead of absolute file system paths, the recorded file nameswill begin either a module path@version (when using modules),or a plain import path (when using the standard library, or gopath)."
    )
    toolexec: list[str] = core.Field(
        default_factory=list,
        cli=cli.List(name="-toolexec"),
        description="A program to use to invoke toolchain programs like vet and asm.For example, instead of running asm, the go command will run'cmd args /path/to/asm <arguments for asm>'.The TOOLEXEC_IMPORTPATH environment variable will be set,matching 'go list -f {{.ImportPath}}' for the package being built."
    )
    source: list[str | Path] = core.Field(
        default_factory=list,
        cli=cli.Args(name="sources"),
        description="Source files / directories / packages."
    )


class Command(core.Model):
    shell: Shell = core.Field(
        default_factory=lambda: Shell(name="go", cmd=["go"]),
        description="Golang tool command."
    )


class Mod(Command):
    @core.typeguard
    def tidy(self, compat: str = "") -> Shell:
        shell = self.shell
        shell.cmd += ["mod", "tidy"]
        if compat.strip():
            shell.cmd.append(f"compat={compat.strip()}")
        return shell

    @core.typeguard
    def vendor(self, outdir: None | Path = None):
        shell = self.shell
        shell.cmd += ["mod", "vendor"]
        if outdir:
            out = Path(outdir).expanduser().resolve().absolute()
            shell.cmd.append(f"-o={out}")
        return shell


class Go:
    @property
    def shell(self) -> Shell:
        return self._shell

    @shell.setter
    def shell(self, value: Shell):
        self._shell = value

    def __init__(self):
        self._shell = Shell(name="go", cmd=["go"])
        self._mod = Mod()

    @property
    def mod(self) -> Mod:
        result = copy.deepcopy(self._mod)
        result.shell = copy.deepcopy(self._shell)
        return result

    def build(self, options: BuildOptions):
        opt = options.serialize()
        shell = copy.deepcopy(self._shell)
        shell.cmd.append("build")
        shell.cmd += opt
        shell.sync()



# Docker
from python_on_whales import Builder
from python_on_whales import Config
from python_on_whales import Container
from python_on_whales import ContainerStats
from python_on_whales import Context
from python_on_whales import DockerClient as Client
from python_on_whales import DockerContextConfig as ContextConfig
from python_on_whales import DockerException
from python_on_whales import Image
from python_on_whales import KubernetesContextConfig
from python_on_whales import Network
from python_on_whales import Node
from python_on_whales import Plugin
from python_on_whales import Secret
from python_on_whales import Service
from python_on_whales import Stack
from python_on_whales import SystemInfo
from python_on_whales import Task
from python_on_whales import Volume

# Dockerfile
from .file import File
from .file import Add as FileAdd
from .file import Arg as FileArg
from .file import Cmd as FileCmd
from .file import Copy as FileCopy
from .file import Entrypoint as FileEntrypoint
from .file import Env as FileEnv
from .file import Expose as FileExpose
from .file import From as FileFrom
from .file import Healthcheck as FileHealthcheck
from .file import Label as FileLabel
from .file import Maintainer as FileMaintainer
from .file import OnBuild as FileOnBuild
from .file import Run as FileRun
from .file import Shell as FileShell
from .file import User as FileUser
from .file import Volume as FileVolume
from .file import Workdir as FileWorkdir

# Compose
from .compose import BlockIo as ComposeBlockIo
from .compose import Build as ComposeBuild
from .compose import Config as ComposeConfig
from .compose import Credential as ComposeCredential
from .compose import Dependency as ComposeDependency
from .compose import Deploy as ComposeDeploy
from .compose import Develop as ComposeDevelop
from .compose import Device as ComposeDevice
from .compose import EnvFile as ComposeEnvFile
from .compose import Extend as ComposeExtend
from .compose import File as ComposeFile
from .compose import Healthcheck as ComposeHealthcheck
from .compose import IPAM as ComposeIpam
from .compose import IpamConfig as ComposeIpamConfig
from .compose import Logging as ComposeLogging
from .compose import Mount as ComposeMount
from .compose import Net as ComposeNet
from .compose import Network as ComposeNetwork
from .compose import Placement as ComposePlacement
from .compose import Resource as ComposeResource
from .compose import Resources as ComposeResources
from .compose import Restart as ComposeRestart
from .compose import Secret as ComposeSecret
from .compose import SecretAccess as ComposeSecretAccess
from .compose import Service as ComposeService
from .compose import StorageOpt as ComposeStorageOpt
from .compose import ULimits as ComposeULimits
from .compose import Update as ComposeUpdate
from .compose import Volume as ComposeVolume
from .compose import Volumes as ComposeVolumes
from .compose import Watch as ComposeWatch


__all__ = [
    "Client",
    "Builder",
    "Config",
    "Container",
    "ContainerStats",
    "Context",
    "ContextConfig",
    "DockerException",
    "Image",
    "KubernetesContextConfig",
    "Network",
    "Node",
    "Plugin",
    "Secret",
    "Service",
    "Stack",
    "SystemInfo",
    "Task",
    "Volume",

    "File",
    "FileAdd",
    "FileArg",
    "FileCmd",
    "FileCopy",
    "FileEntrypoint",
    "FileEnv",
    "FileExpose",
    "FileFrom",
    "FileHealthcheck",
    "FileLabel",
    "FileMaintainer",
    "FileOnBuild",
    "FileRun",
    "FileShell",
    "FileUser",
    "FileVolume",
    "FileWorkdir",

    "ComposeBlockIo",
    "ComposeBuild",
    "ComposeConfig",
    "ComposeCredential",
    "ComposeDependency",
    "ComposeDeploy",
    "ComposeDevelop",
    "ComposeDevice",
    "ComposeEnvFile",
    "ComposeExtend",
    "ComposeFile",
    "ComposeHealthcheck",
    "ComposeIpam",
    "ComposeIpamConfig",
    "ComposeLogging",
    "ComposeMount",
    "ComposeNet",
    "ComposeNetwork",
    "ComposePlacement",
    "ComposeResource",
    "ComposeResources",
    "ComposeRestart",
    "ComposeSecret",
    "ComposeSecretAccess",
    "ComposeService",
    "ComposeStorageOpt",
    "ComposeULimits",
    "ComposeUpdate",
    "ComposeVolume",
    "ComposeVolumes",
    "ComposeWatch",
]



import functools

import asyncclick

import umk.core.globals
from umk import state


class options:
    @staticmethod
    def remote(func):
        # @asyncclick.option(cls=RemoteOption)
        @asyncclick.option("-R", is_flag=True, flag_value="___umk_flag___", help="Execute in the default or specific remote environment")
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper

    @staticmethod
    def style(func):
        @asyncclick.option("-s", default="style", type=asyncclick.Choice(["style", "json"], case_sensitive=False), help="Output format")
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper

    class config:
        @staticmethod
        def all(func):
            @asyncclick.option("-F", is_flag=True, help="Load config from file")
            @asyncclick.option("-P", required=False, type=str, multiple=True, help="Config preset to apply")
            @asyncclick.option("-C", required=False, type=str, multiple=True, help="Config entry override")
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper

        @staticmethod
        def entry(func):
            @asyncclick.option("-C", required=False, type=str, multiple=True, help="Config entry override")
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper

        @staticmethod
        def preset(func):
            @asyncclick.option("-P", required=False, type=str, multiple=True, help="Config preset to apply")
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper

        @staticmethod
        def file(func):
            @asyncclick.option("-F", is_flag=True, help="Load config from file")
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper


if not state.complete:
    import sys

    from umk import core
    from umk import runtime


    def forward(container: runtime.Container):
        if state.remote.default:
            re = container.find_remote(state.remote.default, "")
        elif state.remote.specific.strip():
            re = container.find_remote(False, state.remote.specific)
        else:
            return
        umk.core.globals.console.print(f"[bold]Forward execution to '{re.name}' remote environment: '{' '.join(state.remote.cmd)}'")
        re.execute(state.remote.cmd)
        sys.exit(0)

    def config(file: bool, presets: tuple[str] | None = None, overrides: None | tuple[str] = None) -> runtime.Options.Config:
        result = runtime.Options.Config()
        result.presets = list(presets) if presets else []
        result.file = file
        if not overrides:
            return result
        if overrides is not None:
            for entry in overrides:
                name = ""
                i = 0
                for i, sym in enumerate(entry):
                    if sym != "=":
                        name += sym
                    # TODO Validate entry symbols
                    else:
                        break
                    if i > len(entry):
                        # TODO Invalid syntax
                        print(f"Invalid config entry: {entry}", file=sys.stderr)
                        core.globals.close(-1)
                value = entry[i+1:]
                result.overrides[name] = value
        return result



from umk import core
from umk.framework.utils import cli
from umk.framework.filesystem import Path, AnyPath
from umk.framework.system.shell import Shell
from umk.framework.system.environs import Environs


class Options(cli.Options):
    accept_multiclient: None | bool = core.Field(
        default=None,
        cli=cli.Bool(name="--accept-multiclient"),
        description="Allows a headless server to accept multiple client connections via JSON-RPC or DAP."
    )
    allow_non_terminal_interactive: None | bool = core.Field(
        default=None,
        cli=cli.Bool(name="--allow-non-terminal-interactive"),
        description="Allows interactive sessions of Delve that don't have a terminal as stdin, stdout and stderr"
    )
    api_version: None | int = core.Field(
        default=None,
        cli=cli.Int(name="--api-version"),
        description="Selects JSON-RPC API version when headless. New clients should use v2. Can be reset via RPCServer.SetApiVersion."
    )
    backend: None | str = core.Field(
        default=None,
        cli=cli.Str(name="--backend"),
        description="Backend selection (see 'dlv help backend')."
    )
    build_flags: None | str = core.Field(
        default=None,
        cli=cli.Str(name="--build-flags"),
        description="Build flags, to be passed to the compiler."
    )
    check_go_version: None | bool = core.Field(
        default=None,
        cli=cli.Bool(name="--check-go-version"),
        description="Exits if the version of Go in use is not compatible (too old or too new) with the version of Delve."
    )
    disable_aslr: None | bool = core.Field(
        default=None,
        cli=cli.Bool(name="--disable-aslr"),
        description="Disables address space randomization"
    )
    headless: None | bool = core.Field(
        default=None,
        cli=cli.Bool(name="--headless"),
        description="Run debug server only, in headless mode. Server will accept both JSON-RPC or DAP client connections."
    )
    init: None | str = core.Field(
        default=None,
        cli=cli.Str(name="--init"),
        description="Init file, executed by the terminal client."
    )
    listen: None | str = core.Field(
        default=None,
        cli=cli.Str(name="--listen"),
        description="Debugging server listen address (default 127.0.0.1:0)."
    )
    log: None | bool = core.Field(
        default=None,
        cli=cli.Bool(name="--log"),
        description="Enable debugging server logging."
    )
    log_dest: None | str = core.Field(
        default=None,
        cli=cli.Str(name="--log-dest"),
        description="Writes logs to the specified file or file descriptor."
    )
    log_output: None | str = core.Field(
        default=None,
        cli=cli.Str(name="--log-output"),
        description="Comma separated list of components that should produce debug output."
    )
    only_same_user: None | bool = core.Field(
        default=None,
        cli=cli.Bool(name="--only-same-user"),
        description="Only connections from the same user that started this instance of Delve are allowed to connect."
    )
    redirect: dict[str, str] = core.Field(
        default_factory=dict,
        cli=cli.Dict(name="--redirect"),
        description="Specifies redirect rules for target process."
    )
    wd: None | str = core.Field(
        default=None,
        cli=cli.Str(name="--wd"),
        description="Working directory for running the program."
    )


class Delve(core.Model):
    cmd: list[AnyPath] = core.Field(
        default_factory=lambda: ["dlv"],
        description="Delve command."
    )
    options: Options = core.Field(
        default_factory=Options,
        description="Delve main options."
    )
    workdir: Path = core.Field(
        default_factory=lambda: core.globals.paths.work,
        description="Working directory."
    )
    environs: None | Environs = core.Field(
        default=None,
        description="Shell environment variables."
    )

    @property
    def shell(self) -> Shell:
        result = Shell(name="delve")
        result.cmd = self.cmd
        result.cmd += self.options.serialize()
        result.environs = self.environs
        return result

    @core.typeguard
    def attach(self, pid: int, executable: AnyPath = '', continues: bool = False) -> Shell:
        shell = self.shell
        shell.cmd += ["attach", str(pid)]
        if executable:
            shell.cmd.append(executable)
        if continues:
            shell.cmd.append("--continue")
        return shell

    @core.typeguard
    def exec(self, cmd: list[AnyPath], continues: bool = False, tty: str = "") -> Shell:
        shell = self.shell
        shell.cmd += ["exec"]
        if tty:
            shell.cmd += [f"--tty={tty}"]
        if continues:
            shell.cmd.append("--continue")
        shell.cmd.append(cmd[0])
        if len(cmd) > 1:
            shell.cmd.append("--")
            self.cmd += cmd[1:]
        return shell



import os

from umk import core


class EnvironmentNotExistsError(Exception):
    def __init__(self, name: str, message: str = ""):
        self.name = name
        self.message = message

    def __str__(self):
        if self.message:
            return f"Environment '{self.name}' is required. {self.message}"
        else:
            return f"Environment '{self.name}' is required"


class Environs(dict):
    @core.typeguard
    def __init__(self, inherit=True, **var):
        super().__init__()
        if inherit:
            for name, value in os.environ.items():
                self[name] = value
        for k, v in var.items():
            self[k] = v

    @core.typeguard
    def prepend(self, name: str, *values: str) -> None:
        nonempty = filter(None, values)
        value = os.pathsep.join(nonempty)
        if name in self and self[name]:
            value += os.pathsep + self[name]
        self[name] = value

    @core.typeguard
    def append(self, name: str, *values: str) -> None:
        nonempty = filter(None, values)
        appended = os.pathsep.join(nonempty)
        if name in self and self[name]:
            appended = f'{self[name]}{os.pathsep}{appended}'
        self[name] = appended

    @core.typeguard
    def require(self, name: str, message: str = ""):
        if name in self:
            return
        raise EnvironmentNotExistsError(name=name, message=message)


OptEnv = None | Environs



import logging
import os

import paramiko
from paramiko import util as paramiko_util

from umk import core
from umk.framework.filesystem import Path
from umk.framework.remote.interface import Interface
from umk.framework.system.environs import Environs
from umk.framework.system.shell import Shell


class SecureShell(Interface):
    host: str = core.Field(
        default="",
        description="Remote server IP address"
    )
    port: int = core.Field(
        default=22,
        description="Remote server port"
    )
    password: str = core.Field(
        default="",
        description="Remote server user password"
    )
    username: str = core.Field(
        default="",
        description="Remote server user name"
    )
    sh: str = core.Field(
        default="",
        description="Default shell (bash, sh, zsh ...)"
    )

    def client(self) -> paramiko.SSHClient:
        paramiko_util.get_logger('paramiko').setLevel(logging.ERROR)
        result = paramiko.SSHClient()
        result.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        result.connect(self.host, self.port, self.username, self.password)
        return result

    def shell(self, **kwargs):
        paramiko_util.get_logger('paramiko').setLevel(logging.ERROR)
        cmd = ["ssh", f"{self.username}@{self.host}", "-p", str(self.port)]
        if self.sh:
            cmd.extend(["-t", self.sh.strip()])
        Shell(name=self.name, cmd=cmd).sync()

    @core.typeguard
    def execute(self, cmd: list[str], cwd: None | Path | str = None, env: None | Environs = None, **kwargs):
        paramiko_util.get_logger('paramiko').setLevel(logging.ERROR)
        client = self.client()
        _, out, err = client.exec_command(
            command=Shell.stringify(cmd),
            environment=None,
            get_pty=True
        )
        for line in out:
            print(line.rstrip())
        for line in err:
            print(line.rstrip())

    @core.typeguard
    def upload(self, paths: dict[str | Path, str | Path], **kwargs):
        if not paths:
            return
        client = self.client()
        with client.open_sftp() as transport:
            for src, dst in paths.items():
                core.globals.console.print(f"[bold]\[{self.name}] upload: {src} -> {dst}")
                transport.put(
                    localpath=str(Path(src).expanduser().resolve().absolute()),
                    remotepath=str(dst)
                )
        client.close()

    @core.typeguard
    def download(self, items: dict[str | Path, str | Path], **kwargs):
        if not items:
            return
        client = self.client()
        with client.open_sftp() as transport:
            for src, dst in items.items():
                core.globals.console.print(f"[bold]\[{self.name}] download: {src} -> {dst}")
                dst = Path(dst).expanduser().resolve().absolute()
                if not dst.parent.exists():
                    os.makedirs(dst.parent)
                transport.get(
                    remotepath=str(src),
                    localpath=str(dst)
                )
        client.close()



from umk import core
from umk.kit import project, config, remote
from umk.runtime import utils


class Remote(core.Model):
    class Decorators(core.Model):
        custom: utils.Decorator = core.Field(
            description="Decorator of the remote 'custom'",
            default_factory=lambda: utils.Decorator(
                stack=2,
                module="remote",
                input=utils.Decorator.Input(
                    subject="class",
                    base=remote.Interface,
                ),
                errors=utils.Decorator.OnErrors(
                    module=utils.SourceError("Failed to register custom remote environment outside of the .unimake/remote.py"),
                    subject=utils.ClassError("Failed to register custom remote environment. Use 'umk.framework.remote.custom' with classes based on 'umk.framework.remote.Interface'"),
                    base=utils.SubclassError("Failed to register custom remote environment. Use 'umk.framework.remote.custom' with classes based on 'umk.framework.remote.Interface'"),
                )
            )
        )
        ssh: utils.Decorator = core.Field(
            description="Decorator of the remote 'ssh'",
            default_factory=lambda: utils.Decorator(
                stack=2,
                input=utils.Decorator.Input(
                    subject="function",
                    sig=utils.Decorator.Input.Signature(min=1)
                ),
                module="remote",
                errors=utils.Decorator.OnErrors(
                    module=utils.SourceError("Failed to register remote environment 'ssh' outside of the .unimake/remote.py"),
                    subject=utils.FunctionError("Failed to register remote environment 'ssh'. Use 'umk.framework.remote.ssh with functions"),
                    sig=utils.SignatureError("Failed to register remote environment 'ssh'. Function must accept 1 argument at least"),
                )
            ),
        )
        container: utils.Decorator = core.Field(
            description="Decorator of the remote 'docker.container'",
            default_factory=lambda: utils.Decorator(
                stack=2,
                input=utils.Decorator.Input(
                    subject="function",
                    sig=utils.Decorator.Input.Signature(min=1)
                ),
                module="remote",
                errors=utils.Decorator.OnErrors(
                    module=utils.SourceError("Failed to register remote environment 'docker.container' outside of the .unimake/remote.py"),
                    subject=utils.FunctionError("Failed to register remote environment 'docker.container'. Use 'umk.framework.remote.docker.container with functions"),
                    sig=utils.SignatureError("Failed to register remote environment 'docker.container'. Function must accept 1 argument at least"),
                )
            ),
        )
        compose: utils.Decorator = core.Field(
            description="Decorator of the remote 'docker.compose'",
            default_factory=lambda: utils.Decorator(
                stack=2,
                input=utils.Decorator.Input(
                    subject="function",
                    sig=utils.Decorator.Input.Signature(min=1)
                ),
                module="remote",
                errors=utils.Decorator.OnErrors(
                    module=utils.SourceError("Failed to register remote environment 'docker.compose' outside of the .unimake/remote.py"),
                    subject=utils.FunctionError("Failed to register remote environment 'docker.compose'. Use 'umk.framework.remote.docker.compose with functions"),
                    sig=utils.SignatureError("Failed to register remote environment 'docker.compose'. Function must accept 1 argument at least"),
                )
            ),
        )
    decorator: Decorators = core.Field(
        default_factory=Decorators,
        description="Targets decorators"
    )
    items: dict[str, remote.Interface] = core.Field(
        default_factory=dict,
        description="List of the remote environments"
    )
    default: str = core.Field(
        default="",
        description="Default remote environment name"
    )

    def __iter__(self):
        for rem in self.items.values():
            yield rem

    def __getitem__(self, name: str):
        return self.items[name]

    def __len__(self):
        return len(self.items)

    def __contains__(self, name: str):
        return name in self.items

    def get(self, name: str, on_err=None) -> remote.Interface:
        return self.items.get(name, on_err)

    def find(self, name: str, on_err=None) -> remote.Interface | None:
        return self.get(name, on_err)

    def init(self):
        remote.iterate = self.__iter__
        remote.find = self.find
        remote.default = lambda on_err=None: self.get(self.default, on_err)
        remote.ssh = self.decorator.ssh.register
        remote.docker.container = self.decorator.container.register
        remote.docker.compose = self.decorator.compose.register
        remote.custom = self.decorator.custom.register

    def setup(self, c: config.Interface, p: project.Interface):
        def append(r: remote.Interface):
            if r.name in self.items:
                e = utils.ExistsError()
                e.messages.append(f"Remote environment '{r.name}' is already registered")
                e.details.new(name="name", value=r.name, desc="Remote environment name")
                raise e
            if r.default:
                if self.default:
                    e = utils.ExistsError()
                    e.messages = f"Remote environment '{r.name}' is marked as default, but default is already exists: '{self.default}'"
                    raise e
                self.default = r.name
            self.items[r.name] = r
        for defer in self.decorator.compose.defers:
            src = remote.DockerCompose()
            sig = self.decorator.compose.input.sig
            defer(sig.min, sig.max, src, c, p)
            append(src)
        for defer in self.decorator.container.defers:
            src = remote.DockerContainer()
            sig = self.decorator.container.input.sig
            defer(sig.min, sig.max, src, c, p)
            append(src)
        for defer in self.decorator.ssh.defers:
            src = remote.SecureShell()
            sig = self.decorator.ssh.input.sig
            defer(sig.min, sig.max, src, c, p)
            append(src)
        for defer in self.decorator.custom.defers:
            res = defer(0, 2, c, p)
            append(res)



# base
from umk.framework.project.base import Contributor
from umk.framework.project.base import Info
from umk.framework.project.base import Layout
from umk.framework.project.base import Interface
from umk.framework.project.base import Scratch

# golang
from umk.framework.project.golang import Golang
from umk.framework.project.golang import GolangLayout


def custom(func):
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


def empty(func):
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


def golang(func):
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


def releaser(func):
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


def release():
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


def get():
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()



import copy

from umk import core
from umk.framework.target.interface import Interface
from umk.framework.system.shell import Shell
from umk.framework.system.platform import platform
from shutil import which


class Manager(core.Model):
    binary: str = core.Field(
        default="",
        description="Package manager binary"
    )
    before: Shell = core.Field(
        default_factory=Shell,
        description="Command to run before packages installation"
    )
    install: Shell = core.Field(
        default_factory=Shell,
        description="Command to install packages"
    )
    after: Shell = core.Field(
        default_factory=Shell,
        description="Command to run after packages installation"
    )

    def object(self) -> core.Object:
        result = core.Object()
        result.type = "SystemPackageManager"
        result.properties.new("Binary", self.binary, "Package manager binary")
        result.properties.new("Before", self.before.cmd, "Package manager 'before' command")
        result.properties.new("After", self.after.cmd, "Package manager 'after' command")
        result.properties.new("Install", self.install.cmd, "Package manager 'install' command")
        return result


class PackageList(core.Model):
    manager: Manager = core.Field(
        default=None,
        description="Package manager commands",
    )
    items: list[str] = core.Field(
        default_factory=list,
        description="List ot the packages to install"
    )
    sudo: bool = core.Field(
        default=False,
        description="Run shell by super user"
    )

    def install(self, **kwargs):
        if not self.items:
            core.globals.console.print(
                f"[bold cyan]Skip '{self.manager.binary}' packages, list is empty"
            )
            return

        # set sudo if required
        manager = copy.deepcopy(self.manager)
        if self.sudo and platform().unix:
            if manager.before.cmd:
                manager.before.cmd.insert(0, "sudo")
            if manager.after.cmd:
                manager.after.cmd.insert(0, "sudo")
            if manager.install.cmd:
                manager.install.cmd.insert(0, "sudo")

        # run 'before' command
        if manager.before.cmd:
            manager.before.sync()

        # install packages
        manager.install.cmd += self.items
        manager.install.sync()

        # run 'after' command
        if manager.after.cmd:
            manager.after.sync()

    def object(self) -> core.Object:
        result = core.Object()
        result.type = "SystemPackageList"
        result.properties.new("Sudo", self.sudo, "Run manager commands by super user")
        result.properties.new("Manager", self.manager.object(), "Package manager info")
        result.properties.new("Items", self.items, "Package list")
        return result


class SystemPackages(Interface):
    apt: PackageList = core.Field(
        description="Apt package manager",
        default_factory=lambda: PackageList(
            manager=Manager(
                binary="apt",
                before=Shell(name="apt", cmd=["apt", "-y", "update"]),
                install=Shell(name="apt", cmd=["apt", "-y", "install"]),
                after=Shell(name="apt:clean", cmd=["rm", "-r", "-f", "/var/lib/apt/lists/*"]),
            )
        )
    )
    apt_get: PackageList = core.Field(
        description="Apt-get package manager",
        default_factory=lambda: PackageList(
            manager=Manager(
                binary="apt-get",
                before=Shell(name="apt-get", cmd=["apt-get", "-y", "update"]),
                install=Shell(name="apt-get", cmd=["apt-get", "-y", "install"]),
                after=Shell(name="apt-get:clean", cmd=["rm", "-r", "-f", "/var/lib/apt/lists/*"]),
            )
        )
    )
    apk: PackageList = core.Field(
        description="Apk package manager",
        default_factory=lambda: PackageList(
            manager=Manager(
                binary="apk",
                before=Shell(name="apk", cmd=["apk", "update"]),
                install=Shell(name="apk", cmd=["apk", "add"]),
            )
        )
    )
    yum: PackageList = core.Field(
        description="Yum package manager",
        default_factory=lambda: PackageList(
            manager=Manager(
                binary="yum",
                before=Shell(name="yum", cmd=["yum", "-y", "update"]),
                install=Shell(name="yum", cmd=["yum", "-y", "install"]),
            )
        )
    )
    dnf: PackageList = core.Field(
        description="Dnf package manager",
        default_factory=lambda: PackageList(
            manager=Manager(
                binary="dnf",
                before=Shell(name="dnf", cmd=["dnf", "-y", "update"]),
                install=Shell(name="dnf", cmd=["dnf", "-y", "install"]),
            )
        )
    )
    pacman: PackageList = core.Field(
        description="Pacman package manager",
        default_factory=lambda: PackageList(
            manager=Manager(
                binary="pacman",
                install=Shell(name="pacman", cmd=["pacman", "-S", "--noconfirm", "--disable-download-timeout"]),
            )
        )
    )

    def run(self, **kwargs):
        managers = ("apt-get", "apt", "apk", "yum", "dnf", "pacman")
        binary = None
        for entry in managers:
            binary = which(entry)
            if binary:
                binary = entry
                break
        if binary is None:
            core.globals.console.print(f"[yellow bold]Failed to install packages: no package manager was found !")
            core.globals.console.print(f"[yellow bold]Found one of: {', '.join(managers)}")
            return
        core.globals.console.print(f"[cyan bold]Found '{binary}' package manager")
        pl: PackageList = getattr(self, binary.replace("-", "_"))
        pl.install()

    def object(self, **kwargs) -> core.Object:
        result = core.Object()
        result.type = "Target.SystemPackages"
        result.properties.new("apt", self.apt.items, "apt packages")
        result.properties.new("apt-get", self.apt_get.items, "apt-get packages")
        result.properties.new("apk", self.apk.items, "apk packages")
        result.properties.new("dnf", self.dnf.items, "dnf packages")
        result.properties.new("yum", self.yum.items, "yum packages")
        result.properties.new("pacman", self.pacman.items, "pacman packages")
        return result






from beartype.typing import Generator

from umk.framework.target.interface import Interface
from umk.framework.target.interface import Function
from umk.framework.target.interface import Command

from umk.framework.target.packages import SystemPackages

from umk.framework.target.golang import GolangBinary
from umk.framework.target.golang import GolangMod


def run(*names: str):
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


def iterate() -> Generator[Interface, None, None]:
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


def custom(klass):
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


def function(name: str = "", label: str = "", description: str = ""):
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


def command(func):
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


class go:
    @staticmethod
    def binary(debug=True):
        # See implementation in runtime.Instance.implementation()
        raise NotImplemented()

    @staticmethod
    def mod(func):
        # See implementation in runtime.Instance.implementation()
        raise NotImplemented()


def packages(func):
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()



from git import Repo as Repository

from umk import core
from umk.framework.filesystem import Path


@core.typeguard
def tag(repo: Repository, on_error: str) -> str:
    result = on_error
    if repo and repo.tags:
        result = repo.tags[0].name
    return result


@core.typeguard
def repository(root: Path = core.globals.paths.work):
    return Repository(root)



import os

import asyncclick

from umk.application.cmd import root
from umk.application import utils

if not os.environ.get('_UMK_COMPLETE', None):
    from rich.table import Table
    from umk import runtime
    from umk import core
    from umk.framework.remote.interface import Interface


@root.group(help="Remote environments management commands",)
@asyncclick.option('-n', default="", help="Remote environment name")
@utils.options.config.all
@asyncclick.pass_context
async def remote(ctx: asyncclick.Context, n: str, c: tuple[str], p: tuple[str], f: bool):
    pass
    opt = runtime.Options()
    opt.config = utils.config(f, p, c)
    runtime.c.load(opt)

    if ctx.invoked_subcommand != "ls":
        ctx.ensure_object(dict)
        ctx.obj["instance"] = runtime.c.find_remote(n == "", n)


@remote.command(help="Build remote environment")
@asyncclick.pass_context
def build(ctx: asyncclick.Context):
    instance: Interface = ctx.obj.get("instance")
    instance.build()


@remote.command(help="Destroy remote environment")
@asyncclick.pass_context
def destroy(ctx: asyncclick.Context):
    instance: Interface = ctx.obj.get("instance")
    instance.destroy()


@remote.command(help="Start remote environment")
@asyncclick.pass_context
def up(ctx: asyncclick.Context):
    instance: Interface = ctx.obj.get("instance")
    instance.up()


@remote.command(help="Stop remote environment")
@asyncclick.pass_context
def down(ctx: asyncclick.Context):
    instance: Interface = ctx.obj.get("instance")
    instance.down()


@remote.command(help="Login remote environment")
@asyncclick.pass_context
def login(ctx: asyncclick.Context):
    instance: Interface = ctx.obj.get("instance")
    instance.login()


@remote.command(help="Open remote environment shell")
@asyncclick.pass_context
def shell(ctx: asyncclick.Context):
    instance: Interface = ctx.obj.get("instance")
    instance.shell()


@remote.command(help='List project remote environments')
def ls():
    table = Table(show_header=True, show_edge=True, show_lines=False)
    table.add_column("Name", justify="left", style="", no_wrap=True)
    table.add_column("Default", justify="center", style="", no_wrap=True)
    table.add_column("Description", justify="left", style="", no_wrap=True)
    for rem in runtime.c.remotes:
        default = ''
        if rem.default:
            default = 'x'
            table.add_row(f"{rem.name}", default, rem.description, style="yellow bold")
        else:
            table.add_row(rem.name, default, rem.description)
    core.globals.console.print(table)
    core.globals.close()


@remote.command(name='exec', help="Execute command in remote environment")
@asyncclick.argument('program', required=True, nargs=1)
@asyncclick.argument('arguments', required=False, nargs=-1)
@asyncclick.pass_context
def execute(ctx: asyncclick.Context, program: str, arguments: tuple[str]):
    cmd = list(arguments)
    cmd.insert(0, program)
    rem: Interface = ctx.obj.get("instance")
    rem.execute(cmd=cmd)


@remote.command(help='Show remote environment details')
@utils.options.style
@asyncclick.pass_context
def inspect(ctx: asyncclick.Context, s: str):
    details: core.Object = ctx.obj.get("instance").object()
    if s == "style" or s == "":
        table = Table(show_header=True, show_edge=True, show_lines=True)
        table.add_column("Name", justify="left", style="", no_wrap=True)
        table.add_column("Description", justify="left", style="", no_wrap=True)
        table.add_column("Value", justify="left", style="", no_wrap=True)
        for prop in details.properties:
            table.add_row(prop.name, prop.description, str(prop.value))
        core.globals.console.print(table)
    elif s == "json":
        core.globals.console.print_json(core.json.text(details))


@remote.command(help="Upload files to remote environment")
@asyncclick.argument('items', required=False, nargs=-1)
@asyncclick.pass_context
def upload(ctx: asyncclick.Context, items: tuple[str]):
    rem: Interface = ctx.obj.get("instance")
    paths = split(items)
    if not paths:
        return
    rem.upload(paths)


@remote.command(help="Download files from remote environment")
@asyncclick.argument('items', required=False, nargs=-1)
@asyncclick.pass_context
def download(ctx: asyncclick.Context, items: tuple[str]):
    rem: Interface = ctx.obj.get("instance")
    paths = split(items)
    if not paths:
        return
    rem.download(paths)


def split(items: tuple[str]) -> dict[str, str]:
    def error():
        core.globals.console.print("[bold red]Invalid upload/download item!")
        core.globals.console.print("[bold red]Pattern")
        core.globals.console.print("[bold red]    upload:   <local/file>:<remote/file>")
        core.globals.console.print("[bold red]    download: <remote/file>:<local/file>")
    result = {}
    for item in items:
        if ':' not in item:
            error()
            return {}
        s = item.split(':')
        result[s[0]] = s[1]
    return result



from umk.core.typings import Model

Value = str | int | bool | float
Interface = Model


def register(klass):
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


def preset(*, name: str = ""):
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()


def get():
    # See implementation in runtime.Instance.implementation()
    raise NotImplemented()



import asyncio

from umk import state
from umk.application import cmd


if state.complete or state.unsafe:
    asyncio.run(cmd.root())
else:
    try:
        asyncio.run(cmd.root())
    except Exception as err:
        from umk import runtime
        runtime.errors(err)



import abc
from collections import OrderedDict

from umk.core.serialization import json
from umk.core.typings import Any
from umk.core.typings import Model, Field
from umk.core.typings import typeguard


class Property(Model):
    name: str = ""
    description: str = ""
    value: Any = None

    @staticmethod
    def from_field(instance: Any, field: str):
        f = instance.model_fields[field]
        return Property(
            name=field,
            description=f.description,
            value=getattr(instance, field)
        )

    def __hash__(self):
        return hash(self.name)


class Properties:
    def __init__(self, *properties: Property):
        self._items: OrderedDict[str, Property] = OrderedDict()
        for prop in properties:
            self._items[prop.name] = prop

    @typeguard
    def __contains__(self, prop: str | Property) -> bool:
        t = type(prop)
        if issubclass(t, str):
            return prop in self._items
        else:
            return prop.name in self._items

    def __iter__(self):
        for prop in self._items.values():
            yield prop

    @typeguard
    def __getitem__(self, name: str) -> Property:
        return self._items[name]

    @typeguard
    def __setitem__(self, name: str, value: Property):
        self._items[name] = value

    def __bool__(self):
        return bool(self._items)

    @typeguard
    def get(self, name: str, value: Any = None) -> Property:
        return self._items.get(name, value)

    @typeguard
    def add(self, prop: Property):
        self._items[prop.name] = prop

    @typeguard
    def new(self, name: str, value: Any, desc: str = ""):
        self._items[name] = Property(name=name, value=value, description=desc)


class Object(Model):
    type: str = Field(
        default="",
        description="Object type name"
    )
    properties: Properties = Field(
        default_factory=Properties,
        description="Object properties"
    )


@json.representer(Properties)
def serializing(properties: Properties):
    result = []
    for prop in properties:
        result.append(prop.model_dump())
    return result


@json.representer(Object)
def serializing(obj: Object):
    return obj.model_dump()



import io
import os
from datetime import timedelta

from umk import core
from umk.core.typings import TextIO, Optional, Any
from umk.framework.system.user import User as OSUser
from umk.framework.filesystem import Path


class Instruction(core.Model):
    def write(self, buffer: TextIO):
        pass

    def __str__(self):
        buf = io.StringIO()
        self.write(buf)
        return buf.getvalue()


class Commentable(Instruction):
    comment: list[str] = core.Field(
        description="Instruction comment lines",
        default_factory=list,
    )
    space: int = core.Field(
        description="Instruction pre-space count",
        default=0,
        exclude=True
    )

    def write(self, buffer: TextIO):
        if self.space or self.comment:
            buffer.write("\n")
        if self.comment:
            for line in self.comment:
                text = f"# {line}\n"
                buffer.write(text)


class Add(Commentable):
    """
    Add local or remote files and directories.
    """
    src: str = ""
    dst: str = ""
    chown: Optional[OSUser] = None
    chmod: Optional[int] = None
    checksum: Optional[str] = None

    def write(self, buffer: TextIO):
        super().write(buffer)
        text = "ADD"
        if self.chown is not None:
            text += f" --chown={self.chown.name}"
            if self.chown.group.name:
                text += f":{self.chown.group.name}"
        if self.chmod is not None:
            text += f" --chmod={self.chmod}"
        if self.checksum is not None:
            text += f" --checksum={self.checksum}"
        if text:
            text += " "
        text += f"{self.src} {self.dst}\n"
        buffer.write(text)


class Arg(Commentable):
    """
    Use build-time variables.
    """
    name: str
    value: Any = None

    def write(self, buffer: TextIO):
        super().write(buffer)
        text = f"ARG {self.name}"
        if self.value is not None:
            text += f"={self.value}"
        text += "\n"
        buffer.write(text)


class Cmd(Commentable):
    """ 
    Specify default commands.
    """
    args: list[str] = core.Field(default_factory=list)

    def write(self, buffer: TextIO):
        super().write(buffer)
        text = f"CMD {self.args}\n"
        buffer.write(text)


class Copy(Commentable):
    """ 
    Copy files and directories.
    """
    src: str = ""
    dst: str = ""
    chown: Optional[OSUser] = None
    chmod: Optional[int] = None

    def write(self, buffer: TextIO):
        super().write(buffer)
        text = "COPY"
        if self.chown is not None:
            text += f" --chown={self.chown.name}"
            if self.chown.group.name:
                text += f":{self.chown.group.name}"
        if self.chmod is not None:
            text += f" --chmod={self.chmod}"
        if text:
            text += " "
        text += f"{self.src} {self.dst}\n"
        buffer.write(text)


class Entrypoint(Commentable):
    """
    Specify default executable.
    """
    args: list[str] = core.Field(default_factory=list)

    def write(self, buffer: TextIO):
        super().write(buffer)
        text = "ENTRYPOINT ["
        for i, arg in enumerate(self.args):
            text += f'"{arg}"' + [", ", ""][int(i == len(self.args) - 1)]
        text += "]\n"
        buffer.write(text)


class Env(Commentable):
    """
    Set environment variables.
    """
    name: str
    value: Any = None

    def write(self, buffer: TextIO):
        super().write(buffer)
        if self.value:
            text = f"ENV {self.name}={self.value}\n"
        else:
            text = f"ENV {self.name}\n"
        buffer.write(text)


class Expose(Commentable):
    """
    Describe which ports your application is listening on.
    """
    port: int
    protocol: Optional[str] = None

    def write(self, buffer: TextIO):
        super().write(buffer)
        text = f"EXPOSE {self.port}"
        if self.protocol:
            text += f"/{self.protocol}"
        text += "\n"
        buffer.write(text)


class From(Commentable):
    """
    Create a new build stage from a base image.
    """
    image: str
    platform: Optional[str] = None
    alias: Optional[str] = None

    def write(self, buffer: TextIO):
        super().write(buffer)
        text = f"FROM"
        if self.platform:
            text += f" --platform={self.platform}"
        text += f" {self.image}"
        if self.alias:
            text += f" AS {self.alias}"
        text += "\n"
        buffer.write(text)


class Healthcheck(Commentable):
    """
    Check a container's health on startup.
    """
    interval: Optional[timedelta] = None
    timeout: Optional[timedelta] = None
    start_period: Optional[timedelta] = None
    start_interval: Optional[timedelta] = None
    retries: Optional[int] = None
    command: list[str] = core.Field(default_factory=list)

    def write(self, buffer: TextIO):
        super().write(buffer)
        text = f"HEALTHCHECK"
        if self.interval:
            text += f" --interval={self.interval.total_seconds()}s"
        if self.timeout:
            text += f" --timeout={self.timeout.total_seconds()}s"
        if self.start_period:
            text += f" --start_period={self.start_period.total_seconds()}s"
        if self.start_interval:
            text += f" --start_interval={self.start_interval.total_seconds()}s"
        if self.retries:
            text += f" --retries={self.retries}"
        cmd = "NONE"
        if self.command:
            cmd = " ".join(self.command)
        text += f" CMD {cmd}\n"
        buffer.write(text)


class Label(Commentable):
    """
    Add metadata to an image.
    """
    items: dict[str, str] = core.Field(default_factory=dict)

    def write(self, buffer: TextIO):
        super().write(buffer)
        text = "LABEL"
        for name, value in self.items.items():
            text += f' "{name}"="{value}"'
        text += "\n"
        buffer.write(text)


class Maintainer(Commentable):
    """
    Specify the author of an image.
    """
    name: str

    def write(self, buffer: TextIO):
        super().write(buffer)
        text = f"MAINTAINER {self.name}\n"
        buffer.write(text)
        pass


class OnBuild(Commentable):
    """
    Specify instructions for when the image is used in a build.
    """
    instruction: Instruction = core.Field(kw_only=False)

    def write(self, buffer: TextIO):
        super().write(buffer)
        buf = io.StringIO()
        self.instruction.write(buf)
        ins = buf.getvalue().lstrip()
        text = f"ONBUILD {ins}\n"
        buffer.write(text)


class Run(Commentable):
    """
    Execute build commands.
    """
    commands: list[str] = core.Field(default_factory=list)
    separate: bool = core.Field(default=False)

    def write(self, buffer: TextIO):
        super().write(buffer)
        if self.separate:
            text = "\n".join([f"RUN {cmd}" for cmd in self.commands])
        else:
            text = "RUN " + " && \\\n    ".join(self.commands) + "\n"
        buffer.write(text)


class Shell(Commentable):
    """
    Set the default shell of an image.
    """
    args: list[str] = core.Field(default_factory=list)

    def write(self, buffer: TextIO):
        super().write(buffer)
        text = f"SHELL {self.args}\n"
        buffer.write(text)


class StopSignal(Commentable):
    """
    Specify the system call signal for exiting a container.
    """
    signal: int

    def write(self, buffer: TextIO):
        super().write(buffer)
        text = f"STOPSIGNAL {self.signal}\n"
        buffer.write(text)


class User(Commentable):
    """
    Set user and group ID.
    """
    user: str
    group: Optional[str] = None

    def write(self, buffer: TextIO):
        super().write(buffer)
        text = f"USER {self.user}"
        if self.group:
            text += f":{self.group}"
        text += "\n"
        buffer.write(text)


class Volume(Commentable):
    """
    Create volume mounts.
    """
    path: str

    def write(self, buffer: TextIO):
        super().write(buffer)
        text = f"VOLUME {self.path}\n"
        buffer.write(text)


class Workdir(Commentable):
    """
    Change working directory.
    """
    path: str

    def write(self, buffer: TextIO):
        super().write(buffer)
        text = f"WORKDIR {self.path}\n"
        buffer.write(text)


class Comment(Instruction):
    """
    Comment lines.
    """
    lines: list[str] = core.Field(
        description="Comment lines",
        default_factory=list,
    )

    def write(self, buffer: TextIO):
        super().write(buffer)
        for line in self.lines:
            buffer.write(f"# {line}\n")


class Space(Instruction):
    """
    Comment lines.
    """
    count: int = core.Field(
        description="Empty lines count",
        default=1,
    )

    def write(self, buffer: TextIO):
        super().write(buffer)
        buffer.write(max(1, self.count) * "\n")


class File(core.Model):
    instructions: core.typings.Adapter[list[Instruction]] = core.Field(
        default_factory=list,
        description="Dockerfile instruction list",
    )
    path: Path = core.Field(
        default_factory=Path,
        description="Dockerfile output directory",
    )
    name: str = core.Field(
        default="Dockerfile",
        description="Dockerfile output name",
    )

    def __repr__(self):
        return f"dockerfile://{str(self.file)}"

    def __str__(self):
        return self.text()

    @property
    def file(self) -> Path:
        if self.path is None:
            raise ValueError("Dockerfile: output path is None")
        return self.path / self.name

    def write(self, buffer: TextIO):
        for instruction in self.instructions:
            instruction.write(buffer)

    def save(self):
        file = self.file
        if not file.parent.exists():
            os.makedirs(file.parent)
        with open(file, "w") as stream:
            self.write(stream)

    def text(self) -> str:
        buf = io.StringIO()
        self.write(buf)
        return buf.getvalue()

    @core.typeguard
    def add(self, src: str, dst: str, chown: None | OSUser = None, chmod: None | int = None,
            checksum: None | str = None, *, space: int = 1, comment: list[str] = None):
        instruction = Add(comment=comment or [], space=space, src=src, dst=dst, chmod=chmod, chown=chown,
                          checksum=checksum)
        self.instructions.append(instruction)

    @core.typeguard
    def arg(self, name: str, value: Any = None, *, space: int = 1, comment: list[str] = None):
        instruction = Arg(comment=comment or [], space=space, name=name, value=value)
        self.instructions.append(instruction)

    @core.typeguard
    def cmd(self, args: list[str], *, space: int = 1, comment: list[str] = None):
        instruction = Cmd(comment=comment or [], space=space, args=args)
        self.instructions.append(instruction)

    @core.typeguard
    def copy(self, src: str, dst: str, chown: None | OSUser = None, chmod: None | int = None, *, space: int = 1,
             comment: list[str] = None):
        instruction = Copy(comment=comment or [], space=space, src=src, dst=dst, chmod=chmod, chown=chown)
        self.instructions.append(instruction)

    @core.typeguard
    def entrypoint(self, args: list[str], *, space: int = 1, comment: list[str] = None):
        instruction = Entrypoint(comment=comment or [], space=space, args=args)
        self.instructions.append(instruction)

    @core.typeguard
    def env(self, name: str, value: Any = None, *, space: int = 1, comment: list[str] = None):
        instruction = Env(comment=comment or [], space=space, name=name, value=value)
        self.instructions.append(instruction)

    @core.typeguard
    def expose(self, port: int, protocol: None | str = None, *, space: int = 1, comment: list[str] = None):
        instruction = Expose(comment=comment or [], space=space, port=port, protocol=protocol)
        self.instructions.append(instruction)

    @core.typeguard
    def froms(self, image: str, platform: None | str = None, alias: None | str = None, *, space: int = 0, comment: list[str] = None):
        instruction = From(comment=comment or [], space=space, image=image, platform=platform, alias=alias)
        self.instructions.append(instruction)

    @core.typeguard
    def healthcheck(
            self,
            interval: None | timedelta = None,
            timeout: None | timedelta = None,
            start_period: None | timedelta = None,
            start_interval: None | timedelta = None,
            retries: None | int = None,
            command: list[str] = None,
            *,
            space: int = 1,
            comment: list[str] = None
    ):
        instruction = Healthcheck(
            comment=comment or [],
            space=space,
            interval=interval,
            timeout=timeout,
            start_period=start_period,
            start_interval=start_interval,
            retries=retries,
            command=command
        )
        self.instructions.append(instruction)

    @core.typeguard
    def label(self, items: dict[str, str], *, space: int = 1, comment: list[str] = None):
        instruction = Label(comment=comment or [], space=space, items=items)
        self.instructions.append(instruction)

    @core.typeguard
    def maintainer(self, name: str, *, space: int = 1, comment: list[str] = None):
        instruction = Maintainer(comment=comment or [], space=space, name=name)
        self.instructions.append(instruction)

    @core.typeguard
    def on_build(self, instruction: Instruction, *, space: int = 1, comment: list[str] = None):
        inst = OnBuild(comment=comment or [], space=space, instruction=instruction)
        self.instructions.append(inst)

    @core.typeguard
    def run(self, commands: list[str], separate: bool = False, *, space: int = 1, comment: list[str] = None):
        instruction = Run(comment=comment or [], separate=separate, space=space, commands=commands)
        self.instructions.append(instruction)

    @core.typeguard
    def shell(self, args: list[str], *, space: int = 1, comment: list[str] = None):
        instruction = Shell(comment=comment or [], space=space, args=args)
        self.instructions.append(instruction)

    @core.typeguard
    def stop_signal(self, signal: int, *, space: int = 1, comment: list[str] = None):
        instruction = StopSignal(comment=comment or [], space=space, signal=signal)
        self.instructions.append(instruction)

    @core.typeguard
    def user(self, user: str | int, group: None | str | int = None, *, space: int = 1, comment: list[str] = None):
        instruction = User(comment=comment or [], space=space, user=str(user), group=str(group) if group else None)
        self.instructions.append(instruction)

    @core.typeguard
    def volume(self, path: str, *, space: int = 1, comment: list[str] = None):
        instruction = Volume(comment=comment or [], space=space, path=path)
        self.instructions.append(instruction)

    @core.typeguard
    def workdir(self, path: str, *, space: int = 1, comment: list[str] = None):
        instruction = Workdir(comment=comment or [], space=space, path=path)
        self.instructions.append(instruction)

    @core.typeguard
    def comment(self, *lines: str):
        instruction = Comment(lines=list(lines))
        self.instructions.append(instruction)

    @core.typeguard
    def space(self, count: int = 1):
        instruction = Space(count=count)
        self.instructions.append(instruction)

    def __iadd__(self, instruction: Instruction):
        self.instructions.append(instruction)
        return self

    def __str__(self) -> str:
        return self.text()

    @core.field.serializer("instructions")
    def _serialize_instructions(self, v: list[Instruction]):
        return [str(i).strip() for i in v]



from umk.framework.system.environs import Environs
from umk.framework.system.environs import OptEnv
from umk.framework.system.shell import Shell
from umk.framework.system.shell import Devnull
from umk.framework.system.shell import Handler as ShellHandler
from umk.framework.system.shell import Colorful as ShellColorful
from umk.framework.system.shell import Fetch as ShellFetch
from umk.framework.system.shell import Std as ShellStd
from umk.framework.system.shell import Devnull as ShellDevnull
from umk.framework.system.user import User
from umk.framework.system.user import user
from shutil import chown
from shutil import which
from umk.framework.system.platform import Platform
from umk.framework.system.platform import platform



import logging
from pathlib import Path

from rich.console import Console
from rich.logging import RichHandler

from umk.core.typings import Any


class Paths:
    def __init__(self, root: Path):
        self.work = root.expanduser().resolve().absolute()
        self.unimake = self.work / '.unimake'
        self.cache = self.unimake / ".cache"
        self.config = self.cache / "config.json"


console = Console()
error_console = Console(stderr=True, style="bold red")
handler = RichHandler(
    rich_tracebacks=True,
    show_time=False,
    console=console,
    markup=False,
)
logging.basicConfig(
    level="INFO",
    format="%(message)s",
    datefmt="[%X]",
    handlers=[handler],
)
log = logging.getLogger("rich")
paths = Paths(Path.cwd())
completion = ''


def print(*objects: Any):
    console.print(*objects)


def close(code=0):
    exit(code)



import json as jsonlib
import os
import pathlib

import yaml as yamllib

from umk.core.typings import Any
from umk.core.typings import Model
from umk.core.typings import typeguard


# ////////////////////////////////////////////////////////////////////////////////////
# YAML
# ////////////////////////////////////////////////////////////////////////////////////
class YamlNoAliasDumper(yamllib.SafeDumper):
    def ignore_aliases(self, data):
        return True

class yaml:
    @staticmethod
    def representer(*types):
        def inner(func):
            def wrapped(dumper, data):
                return func

            for t in types:
                yamllib.add_representer(t, func)
                yamllib.representer.SafeRepresenter.add_representer(t, func)
            return wrapped

        return inner

    @staticmethod
    def text(data: Model | dict[str, Any]) -> str:
        if issubclass(type(data), Model):
            d = data.model_dump()
            return yamllib.dump(d, Dumper=YamlNoAliasDumper)
        return yamllib.dump(data, Dumper=YamlNoAliasDumper)


@yaml.representer(str)
def _yaml_multiline_string(dumper, data):
    if len(data.splitlines()) > 1:  # check for multiline string
        return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
    return dumper.represent_scalar('tag:yaml.org,2002:str', data)


@yaml.representer(pathlib.PosixPath, pathlib.WindowsPath)
def _yaml_pathlib(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', str(data))


# ////////////////////////////////////////////////////////////////////////////////////
# JSON
# ////////////////////////////////////////////////////////////////////////////////////

_json_representers = {}


class _JsonEncoder(jsonlib.JSONEncoder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def default(self, value: Any):
        t = type(value)
        for rt, re in _json_representers.items():
            if issubclass(t, rt):
                return re(value)
        return super().default(value)


class json:
    @staticmethod
    def representer(*types):
        def inner(func):
            def wrapped(data):
                return func
            global _json_representers
            for t in types:
                _json_representers[t] = func
            return wrapped
        return inner

    @staticmethod
    @typeguard
    def text(data: Model | dict | list, indent=None) -> str:
        if issubclass(type(data), Model):
            d = data.model_dump()
            return jsonlib.dumps(d, indent=indent, cls=_JsonEncoder)
        return jsonlib.dumps(data, indent=indent, cls=_JsonEncoder)

    @staticmethod
    @typeguard
    def parse(text: str) -> dict | list:
        return jsonlib.loads(text)

    @staticmethod
    @typeguard
    def load(file: str | pathlib.Path) -> dict | list:
        with open(file, "r") as stream:
            return jsonlib.load(stream)

    @staticmethod
    @typeguard
    def save(data: dict | list | Model, file: str | pathlib.Path, indent=None):
        t = type(data)
        if not file.parent.exists():
            os.makedirs(file.parent)
        with open(file, "w") as stream:
            if issubclass(t, (dict, list)):
                jsonlib.dump(data, stream)
            elif issubclass(t, Model):
                text = json.text(data, indent=indent)
                stream.write(text)


@json.representer(pathlib.PosixPath, pathlib.WindowsPath)
def _json_pathlib(data):
    return str(data)




import os

import asyncclick

from umk.application.cmd import root
from umk.application import utils

if not os.environ.get('_UMK_COMPLETE', None):
    from rich.table import Table
    from umk import runtime, core
    from umk.application.printers import PropertiesPrinter


@root.group(help="Project targets management commands")
@utils.options.config.all
def target(c: tuple[str], p: tuple[str], f: bool):
    opt = runtime.Options()
    opt.config = utils.config(f, p, c)
    runtime.c.load(opt)


@target.command(help="List project targets")
@utils.options.style
def ls(s: str):
    if s == "style":
        table = Table(show_header=True, show_edge=True, show_lines=True)
        table.add_column("Name", justify="left", style="", no_wrap=True)
        table.add_column("Description", justify="left", style="", no_wrap=True)
        for t in runtime.c.targets:
            table.add_row(t.name, t.description)
        core.globals.console.print(table)
    elif s == "json":
        data = [t.object().model_dump() for t in runtime.c.targets]
        core.globals.console.print_json(core.json.text(data))


@target.command(name='inspect', help="Inspect targets details")
@utils.options.style
@asyncclick.argument('names', required=True, nargs=-1)
def inspect(s: str, names: tuple[str]):
    found = []
    for name in names:
        if name not in runtime.c.targets:
            core.globals.console.print(f"[yellow bold]Target '{name}' not found")
        else:
            found.append(name)
    objects = []
    for name in found:
        o = runtime.c.targets.get(name).object()
        objects.append(o)

    if s == "style":
        printer = PropertiesPrinter()
        properties = []
        for o in objects:
            properties.append(o.properties)
        printer.print(*properties)

    elif s == "json":
        core.globals.console.print_json(core.json.text(objects))



import os
import pwd
import grp

from umk import core


class Group(core.Model):
    id: None | int = core.Field(
        default=None,
        description="Group ID."
    )
    name: None | str = core.Field(
        default=None,
        description="Group name."
    )

    def __str__(self):
        return f"gid={self.id} group={self.name}"


class User(Group):
    group: Group = core.Field(
        default_factory=Group,
        description="User group info."
    )

    def __str__(self):
        return f"uid={self.id} user={self.name} {self.group}"


def user() -> User:
    u = pwd.getpwuid(os.getuid())
    g = grp.getgrgid(u.pw_gid)
    return User(
        name=u.pw_name,
        id=u.pw_uid,
        group=Group(name=g.gr_name, id=g.gr_gid)
    )
